# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Thread
from time import sleep

from testing_support.sample_applications import (
    raise_exception_application,
    raise_exception_finalize,
    raise_exception_response,
    simple_app_raw,
)
from testing_support.util import get_open_port


def sample_application(environ, start_response):
    path_info = environ.get("PATH_INFO")

    if path_info.startswith("/raise-exception-application"):
        return raise_exception_application(environ, start_response)
    elif path_info.startswith("/raise-exception-response"):
        return raise_exception_response(environ, start_response)
    elif path_info.startswith("/raise-exception-finalize"):
        return raise_exception_finalize(environ, start_response)

    return simple_app_raw(environ, start_response)


def setup_application():
    port = get_open_port()

    def run_wsgi():
        from waitress import serve

        serve(sample_application, host="127.0.0.1", port=port)

    wsgi_thread = Thread(target=run_wsgi)
    wsgi_thread.daemon = True
    wsgi_thread.start()

    sleep(1)

    return port



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pytest
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics


@pytest.mark.parametrize(
    "endpoint,transaction_name",
    (
        ("/sync", "_target_application:sync"),
        ("/async", "_target_application:non_sync"),
    ),
)
def test_application(caplog, app, endpoint, transaction_name):
    caplog.set_level(logging.ERROR)

    @validate_transaction_metrics(transaction_name, scoped_metrics=[("Function/" + transaction_name, 1)])
    @validate_code_level_metrics(*transaction_name.split(":"))
    def _test():
        response = app.get(endpoint)
        assert response.status == 200

        # Catch context propagation error messages
        assert not caplog.records

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


@pytest.fixture()
def exercise_async_documents(async_collection):
    async def _exercise_async_documents():
        italy_doc = async_collection.document("Italy")
        await italy_doc.set({"capital": "Rome", "currency": "Euro", "language": "Italian"})
        await italy_doc.get()
        italian_cities = italy_doc.collection("cities")
        await italian_cities.add({"capital": "Rome"})
        retrieved_coll = [_ async for _ in italy_doc.collections()]
        assert len(retrieved_coll) == 1

        usa_doc = async_collection.document("USA")
        await usa_doc.create({"capital": "Washington D.C.", "currency": "Dollar", "language": "English"})
        await usa_doc.update({"president": "Joe Biden"})

        await async_collection.document("USA").delete()

    return _exercise_async_documents


def test_firestore_async_documents(loop, exercise_async_documents, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/Italy/set", 1),
        ("Datastore/statement/Firestore/Italy/get", 1),
        ("Datastore/statement/Firestore/Italy/collections", 1),
        ("Datastore/statement/Firestore/cities/add", 1),
        ("Datastore/statement/Firestore/USA/create", 1),
        ("Datastore/statement/Firestore/USA/update", 1),
        ("Datastore/statement/Firestore/USA/delete", 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/set", 1),
        ("Datastore/operation/Firestore/get", 1),
        ("Datastore/operation/Firestore/add", 1),
        ("Datastore/operation/Firestore/collections", 1),
        ("Datastore/operation/Firestore/create", 1),
        ("Datastore/operation/Firestore/update", 1),
        ("Datastore/operation/Firestore/delete", 1),
        ("Datastore/all", 7),
        ("Datastore/allOther", 7),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 7),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_async_documents",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_async_documents")
    def _test():
        loop.run_until_complete(exercise_async_documents())

    _test()


@background_task()
def test_firestore_async_documents_generators(
    collection, async_collection, assert_trace_for_async_generator, instance_info
):
    subcollection_doc = collection.document("SubCollections")
    subcollection_doc.set({})
    subcollection_doc.collection("collection1").add({})
    subcollection_doc.collection("collection2").add({})
    assert len([_ for _ in subcollection_doc.collections()]) == 2

    async_subcollection = async_collection.document(subcollection_doc.id)

    assert_trace_for_async_generator(async_subcollection.collections)


def test_firestore_async_documents_trace_node_datastore_params(loop, exercise_async_documents, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        loop.run_until_complete(exercise_async_documents())

    _test()


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from starlette.applications import Starlette
from starlette.background import BackgroundTasks
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from testing_support.asgi_testing import AsgiTest


class ASGIStyleMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        response = await self.app(scope, receive, send)
        return response


class BaseHTTPStyleMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        # simple middleware that does absolutely nothing
        response = await call_next(request)
        return response


async def run_async_bg_task(request):
    tasks = BackgroundTasks()
    tasks.add_task(async_bg_task)
    return PlainTextResponse("Hello, world!", background=tasks)


async def run_sync_bg_task(request):
    tasks = BackgroundTasks()
    tasks.add_task(sync_bg_task)
    return PlainTextResponse("Hello, world!", background=tasks)


async def async_bg_task():
    pass


def sync_bg_task():
    pass


routes = [
    Route("/async", run_async_bg_task),
    Route("/sync", run_sync_bg_task),
]

# Generating target applications
target_application = {}

app = Starlette(routes=routes)
app.add_middleware(ASGIStyleMiddleware)
target_application["asgi"] = AsgiTest(app)

app = Starlette(routes=routes)
app.add_middleware(BaseHTTPStyleMiddleware)
target_application["basehttp"] = AsgiTest(app)

app = Starlette(routes=routes)
target_application["none"] = AsgiTest(app)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2cffi
import psycopg2cffi.extensions
import psycopg2cffi.extras
from testing_support.db_settings import postgresql_settings
from testing_support.fixtures import validate_stats_engine_explain_plan_output_is_none
from testing_support.util import instance_hostname
from testing_support.validators.validate_database_trace_inputs import (
    validate_database_trace_inputs,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_transaction_slow_sql_count import (
    validate_transaction_slow_sql_count,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

DB_SETTINGS = postgresql_settings()[0]

_test_execute_via_cursor_scoped_metrics = [
    ("Function/psycopg2cffi:connect", 1),
    ("Function/psycopg2cffi._impl.connection:Connection.__enter__", 1),
    ("Function/psycopg2cffi._impl.connection:Connection.__exit__", 1),
    ("Datastore/statement/Postgres/%s/select" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/insert" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/update" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/delete" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/now/call", 1),
    ("Datastore/statement/Postgres/pg_sleep/call", 1),
    ("Datastore/operation/Postgres/drop", 1),
    ("Datastore/operation/Postgres/create", 1),
    ("Datastore/operation/Postgres/commit", 3),
    ("Datastore/operation/Postgres/rollback", 1),
]

_test_execute_via_cursor_rollup_metrics = [
    ("Datastore/all", 13),
    ("Datastore/allOther", 13),
    ("Datastore/Postgres/all", 13),
    ("Datastore/Postgres/allOther", 13),
    ("Datastore/operation/Postgres/select", 1),
    ("Datastore/statement/Postgres/%s/select" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/insert", 1),
    ("Datastore/statement/Postgres/%s/insert" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/update", 1),
    ("Datastore/statement/Postgres/%s/update" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/delete", 1),
    ("Datastore/statement/Postgres/%s/delete" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/drop", 1),
    ("Datastore/operation/Postgres/create", 1),
    ("Datastore/statement/Postgres/now/call", 1),
    ("Datastore/statement/Postgres/pg_sleep/call", 1),
    ("Datastore/operation/Postgres/call", 2),
    ("Datastore/operation/Postgres/commit", 3),
    ("Datastore/operation/Postgres/rollback", 1),
    ("Datastore/instance/Postgres/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]), 12),
]


@validate_transaction_metrics(
    "test_database:test_execute_via_cursor",
    scoped_metrics=_test_execute_via_cursor_scoped_metrics,
    rollup_metrics=_test_execute_via_cursor_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor():
    with psycopg2cffi.connect(
        database=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        password=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
    ) as connection:
        cursor = connection.cursor()

        psycopg2cffi.extensions.register_type(psycopg2cffi.extensions.UNICODE)
        psycopg2cffi.extensions.register_type(psycopg2cffi.extensions.UNICODE, connection)
        psycopg2cffi.extensions.register_type(psycopg2cffi.extensions.UNICODE, cursor)

        cursor.execute("""drop table if exists %s""" % DB_SETTINGS["table_name"])

        cursor.execute("""create table %s """ % DB_SETTINGS["table_name"] + """(a integer, b real, c text)""")

        cursor.executemany(
            """insert into %s """ % DB_SETTINGS["table_name"] + """values (%s, %s, %s)""",
            [(1, 1.0, "1.0"), (2, 2.2, "2.2"), (3, 3.3, "3.3")],
        )

        cursor.execute("""select * from %s""" % DB_SETTINGS["table_name"])

        for row in cursor:
            pass

        cursor.execute(
            """update %s""" % DB_SETTINGS["table_name"] + """ set a=%s, b=%s, """ """c=%s where a=%s""",
            (4, 4.0, "4.0", 1),
        )

        cursor.execute("""delete from %s where a=2""" % DB_SETTINGS["table_name"])

        connection.commit()

        cursor.callproc("now")
        cursor.callproc("pg_sleep", (0,))

        connection.rollback()
        connection.commit()


_test_rollback_on_exception_scoped_metrics = [
    ("Function/psycopg2cffi:connect", 1),
    ("Function/psycopg2cffi._impl.connection:Connection.__enter__", 1),
    ("Function/psycopg2cffi._impl.connection:Connection.__exit__", 1),
    ("Datastore/operation/Postgres/rollback", 1),
]


_test_rollback_on_exception_rollup_metrics = [
    ("Datastore/all", 2),
    ("Datastore/allOther", 2),
    ("Datastore/Postgres/all", 2),
    ("Datastore/Postgres/allOther", 2),
]


@validate_transaction_metrics(
    "test_database:test_rollback_on_exception",
    scoped_metrics=_test_rollback_on_exception_scoped_metrics,
    rollup_metrics=_test_rollback_on_exception_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_rollback_on_exception():
    try:
        with psycopg2cffi.connect(
            database=DB_SETTINGS["name"],
            user=DB_SETTINGS["user"],
            password=DB_SETTINGS["password"],
            host=DB_SETTINGS["host"],
            port=DB_SETTINGS["port"],
        ):
            raise RuntimeError("error")
    except RuntimeError:
        pass


_test_async_mode_scoped_metrics = [
    ("Function/psycopg2cffi:connect", 1),
    ("Datastore/statement/Postgres/%s/select" % DB_SETTINGS["table_name"], 1),
    ("Datastore/statement/Postgres/%s/insert" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/drop", 1),
    ("Datastore/operation/Postgres/create", 1),
]

_test_async_mode_rollup_metrics = [
    ("Datastore/all", 5),
    ("Datastore/allOther", 5),
    ("Datastore/Postgres/all", 5),
    ("Datastore/Postgres/allOther", 5),
    ("Datastore/operation/Postgres/select", 1),
    ("Datastore/statement/Postgres/%s/select" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/insert", 1),
    ("Datastore/statement/Postgres/%s/insert" % DB_SETTINGS["table_name"], 1),
    ("Datastore/operation/Postgres/drop", 1),
    ("Datastore/operation/Postgres/create", 1),
    ("Datastore/instance/Postgres/%s/%s" % (instance_hostname(DB_SETTINGS["host"]), DB_SETTINGS["port"]), 4),
]


@validate_stats_engine_explain_plan_output_is_none()
@validate_transaction_slow_sql_count(num_slow_sql=4)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@validate_transaction_metrics(
    "test_database:test_async_mode",
    scoped_metrics=_test_async_mode_scoped_metrics,
    rollup_metrics=_test_async_mode_rollup_metrics,
    background_task=True,
)
@validate_transaction_errors(errors=[])
@background_task()
def test_async_mode():
    wait = psycopg2cffi.extras.wait_select

    kwargs = {}
    version = get_package_version_tuple("psycopg2cffi")
    assert version is not None
    if version >= (2, 8):
        kwargs["async_"] = 1
    else:
        kwargs["async"] = 1

    async_conn = psycopg2cffi.connect(
        database=DB_SETTINGS["name"],
        user=DB_SETTINGS["user"],
        password=DB_SETTINGS["password"],
        host=DB_SETTINGS["host"],
        port=DB_SETTINGS["port"],
        **kwargs
    )
    wait(async_conn)
    async_cur = async_conn.cursor()

    async_cur.execute("""drop table if exists %s""" % DB_SETTINGS["table_name"])
    wait(async_cur.connection)

    async_cur.execute("""create table %s """ % DB_SETTINGS["table_name"] + """(a integer, b real, c text)""")
    wait(async_cur.connection)

    async_cur.execute("""insert into %s """ % DB_SETTINGS["table_name"] + """values (%s, %s, %s)""", (1, 1.0, "1.0"))
    wait(async_cur.connection)

    async_cur.execute("""select * from %s""" % DB_SETTINGS["table_name"])
    wait(async_cur.connection)

    for row in async_cur:
        assert isinstance(row, tuple)

    async_conn.close()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import pika
import pytest
import six
import tornado
from compat import basic_consume
from conftest import (
    BODY,
    CORRELATION_ID,
    EXCHANGE,
    EXCHANGE_2,
    HEADERS,
    QUEUE,
    QUEUE_2,
    REPLY_TO,
)
from minversion import pika_version_info
from pika.adapters.tornado_connection import TornadoConnection
from testing_support.db_settings import rabbitmq_settings
from testing_support.fixtures import (
    capture_transaction_metrics,
    function_not_called,
    override_application_settings,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


DB_SETTINGS = rabbitmq_settings()[0]

_message_broker_tt_params = {
    'queue_name': QUEUE,
    'routing_key': QUEUE,
    'correlation_id': CORRELATION_ID,
    'reply_to': REPLY_TO,
    'headers': HEADERS.copy(),
}


# Tornado's IO loop is not configurable in versions 5.x and up
try:
    class MyIOLoop(tornado.ioloop.IOLoop.configured_class()):
        def handle_callback_exception(self, *args, **kwargs):
            raise

    tornado.ioloop.IOLoop.configure(MyIOLoop)
except AttributeError:
    pass

connection_classes = [pika.SelectConnection, TornadoConnection]

parametrized_connection = pytest.mark.parametrize('ConnectionClass',
        connection_classes)


_test_select_conn_basic_get_inside_txn_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, 1),
]

if six.PY3:
    _test_select_conn_basic_get_inside_txn_metrics.append(
        (('Function/test_pika_async_connection_consume:'
          'test_async_connection_basic_get_inside_txn.'
          '<locals>.on_message'), 1))
else:
    _test_select_conn_basic_get_inside_txn_metrics.append(
        ('Function/test_pika_async_connection_consume:on_message', 1))


@parametrized_connection
@pytest.mark.parametrize('callback_as_partial', [True, False])
@validate_code_level_metrics("test_pika_async_connection_consume.test_async_connection_basic_get_inside_txn.<locals>", "on_message", py2_namespace="test_pika_async_connection_consume")
@validate_transaction_metrics(
        ('test_pika_async_connection_consume:'
                'test_async_connection_basic_get_inside_txn'),
        scoped_metrics=_test_select_conn_basic_get_inside_txn_metrics,
        rollup_metrics=_test_select_conn_basic_get_inside_txn_metrics,
        background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_get_inside_txn(producer, ConnectionClass,
        callback_as_partial):
    def on_message(channel, method_frame, header_frame, body):
        assert method_frame
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    if callback_as_partial:
        on_message = functools.partial(on_message)

    def on_open_channel(channel):
        channel.basic_get(callback=on_message, queue=QUEUE)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


@parametrized_connection
@pytest.mark.parametrize('callback_as_partial', [True, False])
def test_select_connection_basic_get_outside_txn(producer, ConnectionClass,
        callback_as_partial):
    metrics_list = []

    @capture_transaction_metrics(metrics_list)
    def test_basic_get():
        def on_message(channel, method_frame, header_frame, body):
            assert method_frame
            assert body == BODY
            channel.basic_ack(method_frame.delivery_tag)
            channel.close()
            connection.close()
            connection.ioloop.stop()

        if callback_as_partial:
            on_message = functools.partial(on_message)

        def on_open_channel(channel):
            channel.basic_get(callback=on_message, queue=QUEUE)

        def on_open_connection(connection):
            connection.channel(on_open_callback=on_open_channel)

        connection = ConnectionClass(
                pika.ConnectionParameters(DB_SETTINGS['host']),
                on_open_callback=on_open_connection)

        try:
            connection.ioloop.start()
        except:
            connection.close()
            connection.ioloop.stop()
            raise

    test_basic_get()

    # Confirm that no metrics have been created. This is because no background
    # task should be created for basic_get actions.
    assert not metrics_list


_test_select_conn_basic_get_inside_txn_no_callback_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
]


@pytest.mark.skipif(
    condition=pika_version_info[0] > 0,
    reason='pika 1.0 removed the ability to use basic_get with callback=None')
@parametrized_connection
@validate_transaction_metrics(
    ('test_pika_async_connection_consume:'
            'test_async_connection_basic_get_inside_txn_no_callback'),
    scoped_metrics=_test_select_conn_basic_get_inside_txn_no_callback_metrics,
    rollup_metrics=_test_select_conn_basic_get_inside_txn_no_callback_metrics,
    background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_get_inside_txn_no_callback(producer,
        ConnectionClass):
    def on_open_channel(channel):
        channel.basic_get(callback=None, queue=QUEUE)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


_test_async_connection_basic_get_empty_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
]


@parametrized_connection
@pytest.mark.parametrize('callback_as_partial', [True, False])
@validate_transaction_metrics(
        ('test_pika_async_connection_consume:'
                'test_async_connection_basic_get_empty'),
        scoped_metrics=_test_async_connection_basic_get_empty_metrics,
        rollup_metrics=_test_async_connection_basic_get_empty_metrics,
        background_task=True)
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_get_empty(ConnectionClass,
        callback_as_partial):
    QUEUE = 'test_async_empty'

    def on_message(channel, method_frame, header_frame, body):
        assert False, body.decode('UTF-8')

    if callback_as_partial:
        on_message = functools.partial(on_message)

    def on_open_channel(channel):
        channel.basic_get(callback=on_message, queue=QUEUE)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


_test_select_conn_basic_consume_in_txn_metrics = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
]

if six.PY3:
    _test_select_conn_basic_consume_in_txn_metrics.append(
        (('Function/test_pika_async_connection_consume:'
          'test_async_connection_basic_consume_inside_txn.'
          '<locals>.on_message'), 1))
else:
    _test_select_conn_basic_consume_in_txn_metrics.append(
        ('Function/test_pika_async_connection_consume:on_message', 1))


@parametrized_connection
@validate_transaction_metrics(
        ('test_pika_async_connection_consume:'
                'test_async_connection_basic_consume_inside_txn'),
        scoped_metrics=_test_select_conn_basic_consume_in_txn_metrics,
        rollup_metrics=_test_select_conn_basic_consume_in_txn_metrics,
        background_task=True)
@validate_code_level_metrics("test_pika_async_connection_consume.test_async_connection_basic_consume_inside_txn.<locals>", "on_message", py2_namespace="test_pika_async_connection_consume")
@validate_tt_collector_json(message_broker_params=_message_broker_tt_params)
@background_task()
def test_async_connection_basic_consume_inside_txn(producer, ConnectionClass):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


_test_select_conn_basic_consume_two_exchanges = [
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE, None),
    ('MessageBroker/RabbitMQ/Exchange/Produce/Named/%s' % EXCHANGE_2, None),
    ('MessageBroker/RabbitMQ/Exchange/Consume/Named/%s' % EXCHANGE_2, None),
]

if six.PY3:
    _test_select_conn_basic_consume_two_exchanges.append(
        (('Function/test_pika_async_connection_consume:'
          'test_async_connection_basic_consume_two_exchanges.'
          '<locals>.on_message_1'), 1))
    _test_select_conn_basic_consume_two_exchanges.append(
        (('Function/test_pika_async_connection_consume:'
          'test_async_connection_basic_consume_two_exchanges.'
          '<locals>.on_message_2'), 1))
else:
    _test_select_conn_basic_consume_two_exchanges.append(
        ('Function/test_pika_async_connection_consume:on_message_1', 1))
    _test_select_conn_basic_consume_two_exchanges.append(
        ('Function/test_pika_async_connection_consume:on_message_2', 1))


@parametrized_connection
@validate_transaction_metrics(
        ('test_pika_async_connection_consume:'
                'test_async_connection_basic_consume_two_exchanges'),
        scoped_metrics=_test_select_conn_basic_consume_two_exchanges,
        rollup_metrics=_test_select_conn_basic_consume_two_exchanges,
        background_task=True)
@validate_code_level_metrics("test_pika_async_connection_consume.test_async_connection_basic_consume_two_exchanges.<locals>", "on_message_1", py2_namespace="test_pika_async_connection_consume")
@validate_code_level_metrics("test_pika_async_connection_consume.test_async_connection_basic_consume_two_exchanges.<locals>", "on_message_2", py2_namespace="test_pika_async_connection_consume")
@background_task()
def test_async_connection_basic_consume_two_exchanges(producer, producer_2,
        ConnectionClass):
    global events_received
    events_received = 0

    def on_message_1(channel, method_frame, header_frame, body):
        channel.basic_ack(method_frame.delivery_tag)
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY

        global events_received
        events_received += 1

        if events_received == 2:
            channel.close()
            connection.close()
            connection.ioloop.stop()

    def on_message_2(channel, method_frame, header_frame, body):
        channel.basic_ack(method_frame.delivery_tag)
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY

        global events_received
        events_received += 1

        if events_received == 2:
            channel.close()
            connection.close()
            connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message_1)
        basic_consume(channel, QUEUE_2, on_message_2)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = ConnectionClass(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


# This should not create a transaction
@function_not_called('newrelic.core.stats_engine',
                'StatsEngine.record_transaction')
@override_application_settings({'debug.record_transaction_failure': True})
def test_tornado_connection_basic_consume_outside_transaction(producer):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = TornadoConnection(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise


if six.PY3:
    _txn_name = ('test_pika_async_connection_consume:'
            'test_select_connection_basic_consume_outside_transaction.'
            '<locals>.on_message')
    _test_select_connection_consume_outside_txn_metrics = [
        (('Function/test_pika_async_connection_consume:'
          'test_select_connection_basic_consume_outside_transaction.'
          '<locals>.on_message'), None)]
else:
    _txn_name = (
        'test_pika_async_connection_consume:on_message')
    _test_select_connection_consume_outside_txn_metrics = [
        ('Function/test_pika_async_connection_consume:on_message', None)]


# This should create a transaction
@validate_transaction_metrics(
        _txn_name,
        scoped_metrics=_test_select_connection_consume_outside_txn_metrics,
        rollup_metrics=_test_select_connection_consume_outside_txn_metrics,
        background_task=True,
        group='Message/RabbitMQ/Exchange/%s' % EXCHANGE)
@validate_code_level_metrics("test_pika_async_connection_consume.test_select_connection_basic_consume_outside_transaction.<locals>", "on_message", py2_namespace="test_pika_async_connection_consume")
def test_select_connection_basic_consume_outside_transaction(producer):
    def on_message(channel, method_frame, header_frame, body):
        assert hasattr(method_frame, '_nr_start_time')
        assert body == BODY
        channel.basic_ack(method_frame.delivery_tag)
        channel.close()
        connection.close()
        connection.ioloop.stop()

    def on_open_channel(channel):
        basic_consume(channel, QUEUE, on_message)

    def on_open_connection(connection):
        connection.channel(on_open_callback=on_open_channel)

    connection = pika.SelectConnection(
            pika.ConnectionParameters(DB_SETTINGS['host']),
            on_open_callback=on_open_connection)

    try:
        connection.ioloop.start()
    except:
        connection.close()
        connection.ioloop.stop()
        raise



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pytest
import copy

from newrelic.api.application import application_instance
from newrelic.api.background_task import BackgroundTask
from newrelic.api.transaction import current_transaction
from newrelic.api.asgi_application import asgi_application, ASGIWebTransaction

from testing_support.asgi_testing import AsgiTest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics


distributed_trace_intrinsics = ['guid', 'traceId', 'priority', 'sampled']
inbound_payload_intrinsics = ['parent.type', 'parent.app', 'parent.account',
        'parent.transportType', 'parent.transportDuration']

payload = {
    'v': [0, 1],
    'd': {
        'ac': '1',
        'ap': '2827902',
        'id': '7d3efb1b173fecfa',
        'pa': '5e5733a911cfbc73',
        'pr': 10.001,
        'sa': True,
        'ti': 1518469636035,
        'tr': 'd6b4ba0c3a712ca',
        'ty': 'App',
    }
}
parent_order = ['parent_type', 'parent_account',
                'parent_app', 'parent_transport_type']
parent_info = {
    'parent_type': payload['d']['ty'],
    'parent_account': payload['d']['ac'],
    'parent_app': payload['d']['ap'],
    'parent_transport_type': 'HTTP'
}


@asgi_application()
async def target_asgi_application(scope, receive, send):
    status = '200 OK'
    type = "http.response.start"
    output = b'hello world'
    response_headers = [(b'content-type', b'text/html; charset=utf-8'),
     (b'content-length', str(len(output)).encode('utf-8'))]

    txn = current_transaction()

    # Make assertions on the ASGIWebTransaction object
    assert txn._distributed_trace_state
    assert txn.parent_type == 'App'
    assert txn.parent_app == '2827902'
    assert txn.parent_account == '1'
    assert txn.parent_span == '7d3efb1b173fecfa'
    assert txn.parent_transport_type == 'HTTP'
    assert isinstance(txn.parent_transport_duration, float)
    assert txn._trace_id == 'd6b4ba0c3a712ca'
    assert txn.priority == 10.001
    assert txn.sampled

    await send({
        "type": type,
        "status": status,
        "headers": response_headers,
    })

    await send({
        "type": "http.response.body",
        "body": b"Hello World",
    })

    return [output]


test_application = AsgiTest(target_asgi_application)


_override_settings = {
    'trusted_account_key': '1',
    'distributed_tracing.enabled': True,
}

_metrics = [
        ('Supportability/DistributedTrace/AcceptPayload/Success', 1),
        ('Supportability/TraceContext/Accept/Success', None)
]


@override_application_settings(_override_settings)
@validate_transaction_metrics(
    '',
    group='Uri',
    rollup_metrics=_metrics)
def test_distributed_tracing_web_transaction():
    headers = {'newrelic': json.dumps(payload)}
    response = test_application.make_request('GET', '/', headers=headers)
    assert 'X-NewRelic-App-Data' not in response.headers


class TestAsgiRequest(object):
    scope = {
        'asgi': {'spec_version': '2.1', 'version': '3.0'},
        'client': ('127.0.0.1', 54768),
        'headers': [(b'host', b'localhost:8000')],
        'http_version': '1.1',
        'method': 'GET',
        'path': '/',
        'query_string': b'',
        'raw_path': b'/',
        'root_path': '',
        'scheme': 'http',
        'server': ('127.0.0.1', 8000),
        'type': 'http'
    }

    async def receive(self):
        pass

    async def send(self, event):
        pass


# test our distributed_trace metrics by creating a transaction and then forcing
# it to process a distributed trace payload
@pytest.mark.parametrize('web_transaction', (True, False))
@pytest.mark.parametrize('gen_error', (True, False))
@pytest.mark.parametrize('has_parent', (True, False))
def test_distributed_tracing_metrics(web_transaction, gen_error, has_parent):
    def _make_dt_tag(pi):
        return "%s/%s/%s/%s/all" % tuple(pi[x] for x in parent_order)

    # figure out which metrics we'll see based on the test params
    # note: we'll always see DurationByCaller if the distributed
    # tracing flag is turned on
    metrics = ['DurationByCaller']
    if gen_error:
        metrics.append('ErrorsByCaller')
    if has_parent:
        metrics.append('TransportDuration')

    tag = None
    dt_payload = copy.deepcopy(payload)

    # if has_parent is True, our metric name will be info about the parent,
    # otherwise it is Unknown/Unknown/Unknown/Unknown
    if has_parent:
        tag = _make_dt_tag(parent_info)
    else:
        tag = _make_dt_tag(dict((x, 'Unknown') for x in parent_info.keys()))
        del dt_payload['d']['tr']

    # now run the test
    transaction_name = "test_dt_metrics_%s" % '_'.join(metrics)
    _rollup_metrics = [
        ("%s/%s%s" % (x, tag, bt), 1)
        for x in metrics
        for bt in ['', 'Web' if web_transaction else 'Other']
    ]

    def _make_test_transaction():
        application = application_instance()
        request = TestAsgiRequest()

        if not web_transaction:
            return BackgroundTask(application, transaction_name)

        tn = ASGIWebTransaction(application, request.scope,
                                request.send, request.receive)
        tn.set_transaction_name(transaction_name)
        return tn

    @override_application_settings(_override_settings)
    @validate_transaction_metrics(
        transaction_name,
        background_task=not(web_transaction),
        rollup_metrics=_rollup_metrics)
    def _test():
        with _make_test_transaction() as transaction:
            transaction.accept_distributed_trace_payload(dt_payload)

            if gen_error:
                try:
                    1 / 0
                except ZeroDivisionError:
                    transaction.notice_error()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import pytest
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction
from testing_support.fixtures import (
    collector_agent_registration_fixture,
    collector_available_fixture,
)

_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
    "application_logging.enabled": True,
    "application_logging.forwarding.enabled": True,
    "application_logging.metrics.enabled": True,
    "application_logging.local_decorating.enabled": True,
    "event_harvest_config.harvest_limits.log_event_data": 100000,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (logger_structlog)",
    default_settings=_default_settings,
)


@pytest.fixture(scope="session")
def StructLogCapLog():
    from structlog import PrintLogger

    class _StructLogCapLog(PrintLogger):
        def __init__(self, caplog=None):
            self.caplog = caplog if caplog is not None else []

        def msg(self, event, **kwargs):
            self.caplog.append(event)
            return

        log = debug = info = warn = warning = msg
        fatal = failure = err = error = critical = exception = msg

        def __repr__(self):
            return "<StructLogCapLog %s>" % str(id(self))

        __str__ = __repr__

    return _StructLogCapLog


@pytest.fixture
def set_trace_ids():
    def _set():
        txn = current_transaction()
        if txn:
            txn._trace_id = "abcdefgh12345678"
        trace = current_trace()
        if trace:
            trace.guid = "abcdefgh"

    return _set


def drop_event_processor(logger, method_name, event_dict):
    from structlog import DropEvent

    if method_name == "info":
        raise DropEvent
    else:
        return event_dict


@pytest.fixture(scope="function")
def structlog_caplog(StructLogCapLog):
    return StructLogCapLog()


@pytest.fixture(scope="function")
def logger(structlog_caplog):
    import structlog

    structlog.configure(processors=[], logger_factory=lambda *args, **kwargs: structlog_caplog)
    _logger = structlog.get_logger()
    return _logger


@pytest.fixture(scope="function")
def filtering_logger(structlog_caplog):
    import structlog

    structlog.configure(
        processors=[drop_event_processor], logger_factory=lambda *args, **kwargs: structlog_caplog
    )
    _filtering_logger = structlog.get_logger()
    return _filtering_logger


@pytest.fixture(scope="function")
def callsite_parameter_logger(structlog_caplog):
    import structlog

    structlog.configure(
        processors=[
            structlog.processors.CallsiteParameterAdder(
                [
                    structlog.processors.CallsiteParameter.FILENAME,
                    structlog.processors.CallsiteParameter.FUNC_NAME,
                    structlog.processors.CallsiteParameter.LINENO,
                ],
            ),
            structlog.processors.KeyValueRenderer(),
        ],
        logger_factory=lambda *args, **kwargs: structlog_caplog,
    )

    _callsite_logger = structlog.get_logger()
    return _callsite_logger


@pytest.fixture
def exercise_logging_multiple_lines(set_trace_ids, logger, structlog_caplog):
    def _exercise():
        set_trace_ids()

        logger.msg("Cat", a=42)
        logger.error("Dog")
        logger.critical("Elephant")

        assert len(structlog_caplog.caplog) == 3

        assert "Cat" in structlog_caplog.caplog[0]
        assert "Dog" in structlog_caplog.caplog[1]
        assert "Elephant" in structlog_caplog.caplog[2]

    return _exercise


@pytest.fixture
def exercise_filtering_logging_multiple_lines(set_trace_ids, filtering_logger, structlog_caplog):
    def _exercise():
        set_trace_ids()

        filtering_logger.msg("Cat", a=42)
        filtering_logger.error("Dog")
        filtering_logger.critical("Elephant")

        assert len(structlog_caplog.caplog) == 2

        assert "Cat" not in structlog_caplog.caplog[0]
        assert "Dog" in structlog_caplog.caplog[0]
        assert "Elephant" in structlog_caplog.caplog[1]

    return _exercise


@pytest.fixture
def exercise_logging_single_line(set_trace_ids, logger, structlog_caplog):
    def _exercise():
        set_trace_ids()
        logger.error("A", key="value")
        assert len(structlog_caplog.caplog) == 1

    return _exercise



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction

from testing_support.fixtures import (override_application_settings,
        core_application_stats_engine)
from testing_support.validators.validate_span_events import (
        validate_span_events)


@override_application_settings({'distributed_tracing.enabled': True})
@validate_span_events(count=1)
@background_task(name='test_span_events_dt_enabled')
def test_span_events_dt_enabled():
    transaction = current_transaction()
    transaction._sampled = False


@override_application_settings({'distributed_tracing.enabled': False})
@validate_span_events(count=0)
@background_task(name='test_span_events_dt_disabled')
def test_span_events_dt_disabled():
    transaction = current_transaction()
    transaction._sampled = False


def test_span_stream_is_singleton():
    stats_engine = core_application_stats_engine()
    workarea = stats_engine.create_workarea()

    # The workarea span stream should be equal to the global span stream
    assert stats_engine.span_stream is workarea.span_stream



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from io import BytesIO as IO
except ImportError:
    import StringIO as IO

import webtest
from flask import Flask, Response, send_file
from flask_compress import Compress

from newrelic.api.transaction import get_browser_timing_header

application = Flask(__name__)

compress = Compress()
compress.init_app(application)


@application.route("/compress")
def index_page():
    return "<body>" + 500 * "X" + "</body>"


@application.route("/html_insertion")
def html_insertion():
    return (
        "<!DOCTYPE html><html><head>Some header</head>"
        "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
        "</body></html>"
    )


@application.route("/html_insertion_manual")
def html_insertion_manual():
    header = get_browser_timing_header()
    header = get_browser_timing_header()

    assert header == ""

    return (
        "<!DOCTYPE html><html><head>Some header</head>"
        "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
        "</body></html>"
    )


@application.route("/html_insertion_unnamed_attachment_header")
def html_insertion_unnamed_attachment_header():
    response = Response(
        response="<!DOCTYPE html><html><head>Some header</head>"
        "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
        "</body></html>"
    )
    response.headers.add("Content-Disposition", "attachment")
    return response


@application.route("/html_insertion_named_attachment_header")
def html_insertion_named_attachment_header():
    response = Response(
        response="<!DOCTYPE html><html><head>Some header</head>"
        "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
        "</body></html>"
    )
    response.headers.add("Content-Disposition", 'attachment; filename="X"')
    return response


@application.route("/html_served_from_file")
def html_served_from_file():
    file = IO()
    contents = b"""
    <!DOCTYPE html><html><head>Some header</head>
    <body><h1>My First Heading</h1><p>My first paragraph.</p>
    </body></html>
    """
    file.write(contents)
    file.seek(0)
    return send_file(file, mimetype="text/html")


@application.route("/text_served_from_file")
def text_served_from_file():
    file = IO()
    contents = b"""
    <!DOCTYPE html><html><head>Some header</head>
    <body><h1>My First Heading</h1><p>My first paragraph.</p>
    </body></html>
    """
    file.write(contents)
    file.seek(0)
    return send_file(file, mimetype="text/plain")


_test_application = webtest.TestApp(application)


@application.route("/empty_content_type")
def empty_content_type():
    response = Response(
        response="<!DOCTYPE html><html><head>Some header</head>"
        "<body><h1>My First Heading</h1><p>My first paragraph.</p>"
        "</body></html>",
        mimetype="",
    )
    assert response.mimetype is None
    return response



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.background_task import background_task
from newrelic.api.transaction import set_transaction_name, set_background_task

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics


# test
@validate_transaction_metrics(
        'test_transaction_name:test_transaction_name_default_bt',
        group='Function', background_task=True)
@background_task()
def test_transaction_name_default_bt():
    pass


@validate_transaction_metrics(
        'test_transaction_name:test_transaction_name_default_wt',
        group='Function', background_task=False)
@background_task()
def test_transaction_name_default_wt():
    set_background_task(False)


@validate_transaction_metrics('Transaction', group='Custom',
        background_task=True)
@background_task()
def test_transaction_name_valid_override_bt():
    set_transaction_name('Transaction', group='Custom')


@validate_transaction_metrics('Transaction', group='Function',
        background_task=True)
@background_task()
def test_transaction_name_empty_group_bt():
    set_transaction_name('Transaction', group='')


@validate_transaction_metrics('Transaction', group='Function/Group',
        background_task=True)
@background_task()
def test_transaction_name_leading_slash_on_group_bt():
    set_transaction_name('Transaction', group='/Group')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import time

from testing_support.fixtures import catch_background_exceptions

from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.packages import six


def validate_ml_events(events):
    @function_wrapper
    def _validate_wrapper(wrapped, instance, args, kwargs):
        record_called = []
        recorded_events = []

        @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
        @catch_background_exceptions
        def _validate_ml_events(wrapped, instance, args, kwargs):
            record_called.append(True)
            try:
                result = wrapped(*args, **kwargs)
            except:
                raise
            recorded_events[:] = []
            recorded_events.extend(list(instance._ml_events))

            return result

        _new_wrapper = _validate_ml_events(wrapped)
        val = _new_wrapper(*args, **kwargs)
        assert record_called
        found_events = copy.copy(recorded_events)

        record_called[:] = []
        recorded_events[:] = []

        for expected in events:
            matching_ml_events = 0
            mismatches = []
            for captured in found_events:
                if _check_event_attributes(expected, captured, mismatches):
                    matching_ml_events += 1
            assert matching_ml_events == 1, _event_details(matching_ml_events, found_events, mismatches)

        return val

    return _validate_wrapper


def _check_event_attributes(expected, captured, mismatches):
    assert len(captured) == 2  # [intrinsic, user attributes]

    intrinsics = captured[0]

    if intrinsics["type"] != expected[0]["type"]:
        mismatches.append("key: type, value:<%s><%s>" % (expected[0]["type"], captured[0].get("type", None)))
        return False

    now = time.time()

    if not (isinstance(intrinsics["timestamp"], int) and intrinsics["timestamp"] <= 1000.0 * now):
        mismatches.append("key: timestamp, value:<%s>" % intrinsics["timestamp"])
        return False

    captured_keys = set(six.iterkeys(captured[1]))
    expected_keys = set(six.iterkeys(expected[1]))
    extra_keys = captured_keys - expected_keys

    if extra_keys:
        mismatches.append("extra_keys: %s" % str(tuple(extra_keys)))
        return False

    for key, value in six.iteritems(expected[1]):
        if key in captured[1]:
            captured_value = captured[1].get(key, None)
        else:
            mismatches.append("key: %s, value:<%s><%s>" % (key, value, captured[1].get(key, None)))
            return False

        if value is not None:
            if value != captured_value:
                mismatches.append("key: %s, value:<%s><%s>" % (key, value, captured_value))
                return False

    return True


def _event_details(matching_ml_events, captured, mismatches):
    details = [
        "matching_ml_events=%d" % matching_ml_events,
        "mismatches=%s" % mismatches,
        "captured_events=%s" % captured,
    ]

    return "\n".join(details)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest

from newrelic.common.package_version_utils import get_package_version_tuple
from testing_support.db_settings import redis_settings

from testing_support.fixture.event_loop import event_loop as loop
from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611

try:
    import aioredis

    AIOREDIS_VERSION = get_package_version_tuple("aioredis")
except ImportError:
    import redis.asyncio as aioredis

    # Fake aioredis version to show when it was moved to redis.asyncio
    AIOREDIS_VERSION = (2, 0, 2)


SKIPIF_AIOREDIS_V1 = pytest.mark.skipif(AIOREDIS_VERSION < (2,), reason="Unsupported aioredis version.")
SKIPIF_AIOREDIS_V2 = pytest.mark.skipif(AIOREDIS_VERSION >= (2,), reason="Unsupported aioredis version.")
DB_SETTINGS = redis_settings()[0]


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (datastore_aioredis)",
    default_settings=_default_settings,
    linked_applications=["Python Agent Test (datastore)"],
)


@pytest.fixture(params=("Redis", "StrictRedis"))
def client(request, loop):
    if AIOREDIS_VERSION >= (2, 0):
        if request.param == "Redis":
            return aioredis.Redis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
        elif request.param == "StrictRedis":
            return aioredis.StrictRedis(host=DB_SETTINGS["host"], port=DB_SETTINGS["port"], db=0)
        else:
            raise NotImplementedError()
    else:
        if request.param == "Redis":
            return loop.run_until_complete(aioredis.create_redis("redis://%s:%d" % (DB_SETTINGS["host"], DB_SETTINGS["port"]), db=0))
        elif request.param == "StrictRedis":
            pytest.skip("StrictRedis not implemented.")
        else:
            raise NotImplementedError()

@pytest.fixture(scope="session")
def key():
    return "AIOREDIS-TEST-" + str(os.getpid())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import newrelic.core.config

settings = newrelic.core.config.global_settings

_logger = logging.getLogger(__name__)


RECORDSQL_OFF = 'off'
RECORDSQL_RAW = 'raw'
RECORDSQL_OBFUSCATED = 'obfuscated'

COMPRESSED_CONTENT_ENCODING_DEFLATE = 'deflate'
COMPRESSED_CONTENT_ENCODING_GZIP = 'gzip'

STRIP_EXCEPTION_MESSAGE = ("Message removed by New Relic 'strip_exception_messages' setting")


def set_error_group_callback(callback, application=None):
    """Set the current callback to be used to determine error groups."""
    from newrelic.api.application import application_instance

    if callback is not None and not callable(callback):
        _logger.error("Error group callback must be a callable, or None to unset this setting.")
        return

    # Check for activated application if it exists and was not given.
    application = application_instance(activate=False) if application is None else application

    # Get application settings if it exists, or fallback to global settings object
    _settings = application.settings if application is not None else settings()

    if _settings is None:
        _logger.error("Failed to set error_group_callback in application settings. Report this issue to New Relic support.")
        return

    if _settings.error_collector:
        _settings.error_collector._error_group_callback = callback



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import function_not_called, override_generic_settings
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import BackgroundTask, background_task
from newrelic.api.database_trace import database_trace
from newrelic.api.datastore_trace import datastore_trace
from newrelic.api.external_trace import external_trace
from newrelic.api.function_trace import FunctionTrace, function_trace
from newrelic.api.memcache_trace import memcache_trace
from newrelic.api.message_trace import message_trace
from newrelic.api.time_trace import current_trace
from newrelic.core.config import global_settings
from newrelic.core.trace_cache import trace_cache


@function_trace("waiter3")
async def child():
    pass


async def waiter(asyncio, event, wait):
    with FunctionTrace(name="waiter1", terminal=True):
        event.set()

        # Block until the parent says to exit
        await wait.wait()

    with FunctionTrace(name="waiter2", terminal=True):
        pass

    await child()


async def task(asyncio, trace, event, wait):
    # Test that the trace has been propagated onto this task
    assert current_trace() is trace

    # Start a function trace, this should not interfere with context in the
    # parent task
    await waiter(asyncio, event, wait)


@background_task(name="test_context_propagation")
async def _test(asyncio, schedule, nr_enabled=True):
    trace = current_trace()

    if nr_enabled:
        assert trace is not None
    else:
        assert trace is None

    events = [asyncio.Event() for _ in range(2)]
    wait = asyncio.Event()
    tasks = [schedule(task(asyncio, trace, events[idx], wait)) for idx in range(2)]

    await asyncio.gather(*(e.wait() for e in events))

    # Test that the current trace is still "trace" even though the tasks are
    # active
    assert current_trace() is trace

    # Unblock the execution of the tasks and wait for the tasks to terminate
    wait.set()
    await asyncio.gather(*tasks)

    return trace


@pytest.mark.parametrize("set_loop", (True, False))
@pytest.mark.parametrize(
    "schedule",
    (
        "create_task",
        "ensure_future",
    ),
)
@validate_transaction_metrics(
    "test_context_propagation",
    background_task=True,
    scoped_metrics=(
        ("Function/waiter1", 2),
        ("Function/waiter2", 2),
        ("Function/waiter3", 2),
    ),
)
def test_context_propagation(event_loop, schedule, set_loop):
    import asyncio

    _loop = None
    if set_loop:

        class TestEventLoop(asyncio.SelectorEventLoop):
            def create_task(self, coro, **kwargs):
                return asyncio.tasks.Task(coro, loop=self, **kwargs)

        _loop = event_loop
        event_loop = TestEventLoop()
        asyncio.set_event_loop(event_loop)

    exceptions = []

    def handle_exception(loop, context):
        exceptions.append(context)

    event_loop.set_exception_handler(handle_exception)

    schedule = getattr(asyncio, schedule, None) or getattr(event_loop, schedule)

    # Keep the trace around so that it's not removed from the trace cache
    # through reference counting (for testing)
    _ = event_loop.run_until_complete(_test(asyncio, schedule))

    # The agent should have removed all traces from the cache since
    # run_until_complete has terminated (all callbacks scheduled inside the
    # task have run)
    assert not trace_cache()

    # Assert that no exceptions have occurred
    assert not exceptions, exceptions
    if _loop:
        asyncio.set_event_loop(_loop)


@override_generic_settings(
    global_settings(),
    {
        "enabled": False,
    },
)
@function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
def test_nr_disabled(event_loop):
    import asyncio

    schedule = asyncio.ensure_future

    exceptions = []

    def handle_exception(loop, context):
        exceptions.append(context)

    event_loop.set_exception_handler(handle_exception)

    event_loop.run_until_complete(_test(asyncio, schedule, nr_enabled=False))

    # Assert that no exceptions have occurred
    assert not exceptions, exceptions


@pytest.mark.parametrize(
    "trace",
    [
        function_trace(name="simple_gen"),
        external_trace(library="lib", url="http://foo.com"),
        database_trace("select * from foo"),
        datastore_trace("lib", "foo", "bar"),
        message_trace("lib", "op", "typ", "name"),
        memcache_trace("cmd"),
    ],
)
def test_two_transactions(event_loop, trace):
    """
    Instantiate a coroutine in one transaction and await it in
    another. This should not cause any errors.
    """
    import asyncio

    tasks = []

    ready = asyncio.Event()
    done = asyncio.Event()

    @trace
    async def task():
        pass

    @background_task(name="create_coro")
    async def create_coro():
        for _ in range(2):
            coro = task()
            tasks.append(coro)

        ready.set()
        await done.wait()

    @background_task(name="await")
    async def await_task():
        await ready.wait()
        while len(tasks) > 0:
            task = tasks.pop()
            await task

        done.set()

    afut = asyncio.ensure_future(create_coro(), loop=event_loop)
    bfut = asyncio.ensure_future(await_task(), loop=event_loop)
    event_loop.run_until_complete(asyncio.gather(afut, bfut))


# Sentinel left in cache transaction exited
async def sentinel_in_cache_txn_exited(asyncio, bg):
    event = asyncio.Event()

    with BackgroundTask(application(), "fg") as txn:
        _ = txn.root_span
        task = asyncio.ensure_future(bg(event))

    await event.wait()
    return task


# Trace left in cache, transaction exited
async def trace_in_cache_txn_exited(asyncio, bg):
    event = asyncio.Event()

    with BackgroundTask(application(), "fg"):
        with FunctionTrace("fg") as _:
            task = asyncio.ensure_future(bg(event))

    await event.wait()
    return task


# Trace left in cache, transaction active
async def trace_in_cache_txn_active(asyncio, bg):
    event = asyncio.Event()

    with BackgroundTask(application(), "fg"):
        with FunctionTrace("fg") as _:
            task = asyncio.ensure_future(bg(event))
        await event.wait()

    return task


@pytest.mark.parametrize(
    "fg",
    (
        sentinel_in_cache_txn_exited,
        trace_in_cache_txn_exited,
        trace_in_cache_txn_active,
    ),
)
def test_transaction_exit_trace_cache(event_loop, fg):
    """
    Verifying that the use of ensure_future will not cause errors
    when traces remain in the trace cache after transaction exit
    """
    import asyncio

    exceptions = []

    def handle_exception(loop, context):
        exceptions.append(context)

    async def bg(event):
        with BackgroundTask(application(), "bg"):
            event.set()

    async def handler():
        task = await fg(asyncio, bg)
        await task

    def _test():
        event_loop.set_exception_handler(handle_exception)
        return event_loop.run_until_complete(handler())

    _test()

    # The agent should have removed all traces from the cache since
    # run_until_complete has terminated
    assert not trace_cache()

    # Assert that no exceptions have occurred
    assert not exceptions, exceptions


def test_incomplete_traces_exit_when_root_exits(event_loop):
    """Verifies that child traces in the same task are exited when the root
    exits"""

    import asyncio

    @function_trace(name="child")
    async def child(start, end):
        start.set()
        await end.wait()

    @background_task(name="parent")
    async def parent():
        start = asyncio.Event()
        end = asyncio.Event()
        task = asyncio.ensure_future(child(start, end))
        await start.wait()
        end.set()
        return task

    @validate_transaction_metrics(
        "parent",
        background_task=True,
        scoped_metrics=[("Function/child", 1)],
    )
    def test(loop):
        return loop.run_until_complete(parent())

    task = test(event_loop)
    event_loop.run_until_complete(task)


def test_incomplete_traces_with_multiple_transactions(event_loop):
    import asyncio

    @background_task(name="dummy")
    async def dummy():
        task = asyncio.ensure_future(child(True))
        await end.wait()
        await task

    @function_trace(name="child")
    async def child(running_at_end=False):
        trace = current_trace()
        start.set()
        await end.wait()
        if running_at_end:
            assert current_trace() is trace
        else:
            assert current_trace() is not trace

    @background_task(name="parent")
    async def parent():
        task = asyncio.ensure_future(child())
        await start.wait()
        return task

    @validate_transaction_metrics(
        "parent",
        background_task=True,
        scoped_metrics=[("Function/child", 1)],
    )
    def parent_assertions(task):
        return event_loop.run_until_complete(task)

    @validate_transaction_metrics(
        "dummy",
        background_task=True,
        scoped_metrics=[("Function/child", 1)],
    )
    def dummy_assertions(task):
        return event_loop.run_until_complete(task)

    async def startup():
        return asyncio.Event(), asyncio.Event()

    async def start_dummy():
        dummy_task = asyncio.ensure_future(dummy())
        await start.wait()
        start.clear()
        return dummy_task

    start, end = event_loop.run_until_complete(startup())

    # Kick start dummy transaction (forcing an ensure_future on another
    # transaction)
    dummy_task = event_loop.run_until_complete(start_dummy())

    # start and end a transaction, forcing the child to truncate
    child_task = parent_assertions(parent())

    # Signal to end any in progress tasks
    # * dummy
    # * dummy->child
    # * parent[ended]->child
    end.set()

    # Wait for dummy/parent->child to terminate
    dummy_assertions(dummy_task)
    event_loop.run_until_complete(child_task)


@validate_transaction_metrics("Parent", background_task=True)
def test_transaction_end_on_different_task(event_loop):
    import asyncio

    txn = BackgroundTask(application(), name="Parent")

    async def parent():
        txn.__enter__()
        child_start = asyncio.Event()
        parent_end = asyncio.Event()
        child_task = asyncio.ensure_future(child(child_start, parent_end))
        await child_start.wait()
        parent_end.set()
        return child_task

    async def child(child_start, parent_end):
        child_start.set()
        await parent_end.wait()

        # Calling asyncio.sleep(0) allows the scheduled done callbacks to run
        # done callbacks are scheduled through call_soon
        await asyncio.sleep(0)
        txn.__exit__(None, None, None)

    async def test():
        task = await asyncio.ensure_future(parent())
        await task

    event_loop.run_until_complete(test())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""These tests check two things:

    1. Using a notfound_view will name the transaction after the not_found
       view, not the default_exceptionresponse_view. There can only be one
       notfound_view per Pyramid application, so to test both cases (default
       and user-defined notfound_view), we use 2 separate apps.

    2. Verifying that we don't double wrap the notfound_view. Because
       we instrument Configurator._derive_view(), and because
       Configurator._derive_view() is called twice when adding a
       notfound_view (once in add_notfound_view(), and again in add_view()),
       our instrumentation checks to see if a view has already been
       wrapped before wrapping.

       If our check fails, then the metric '_test_append_slash_app:not_found'
       will have a count of 2.
"""

import pytest
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.common.package_version_utils import get_package_version_tuple

# Defining a `pytestmark` at the module level means py.test will apply
# this `skipif` conditional to all tests in the module.

pytestmark = pytest.mark.skipif(get_package_version_tuple("pyramid") < (1, 3), reason="requires pyramid >= (1, 3)")


def target_application():
    # We need to delay Pyramid application creation because of ordering
    # issues whereby the agent needs to be initialised before Pyramid is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as Pyramid relies on view handlers being
    # at global scope, so import it from a separate module.

    from _test_append_slash_app import _test_application

    return _test_application


_test_append_slash_app_index_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/pyramid.router:Router.__call__", 1),
    ("Function/_test_append_slash_app:home_view", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_append_slash_app:home_view", scoped_metrics=_test_append_slash_app_index_scoped_metrics
)
def test_index():
    application = target_application()
    response = application.get("/")
    response.mustcontain("INDEX RESPONSE")


_test_not_found_append_slash_scoped_metrics = [
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/pyramid.router:Router.__call__", 1),
    ("Function/pyramid.view:AppendSlashNotFoundViewFactory", 1),
    ("Function/_test_append_slash_app:not_found", 1),
]


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "_test_append_slash_app:not_found", scoped_metrics=_test_not_found_append_slash_scoped_metrics
)
def test_not_found_append_slash():
    application = target_application()
    response = application.get("/foo", status=404)
    response.mustcontain("NOT FOUND")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_patch_thread():
    import gevent.monkey

    gevent.monkey.patch_thread()


def test_patch_ssl_recursion():
    import gevent.monkey

    gevent.monkey.patch_all()

    # Delay imports and test creation of contexts from both bundled and installed urllib3
    import urllib3.util.ssl_

    import newrelic.packages.urllib3.util.ssl_

    context1 = urllib3.util.ssl_.create_urllib3_context()
    context2 = newrelic.packages.urllib3.util.ssl_.create_urllib3_context()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sanic import Blueprint, Sanic
from sanic.exceptions import NotFound, SanicException, ServerError
from sanic.handlers import ErrorHandler
from sanic.response import json
from sanic.router import Router
from sanic.views import HTTPMethodView


try:
    # Old style response streaming
    from sanic.response import stream
except ImportError:
    stream = None


class MethodView(HTTPMethodView):
    async def get(self, request):
        return json({"hello": "world"})

    async def post(self, request):
        return json({"hello": "world"})

    async def put(self, request):
        return json({"hello": "world"})

    async def patch(self, request):
        return json({"hello": "world"})

    async def delete(self, request):
        return json({"hello": "world"})


class CustomErrorHandler(ErrorHandler):
    def response(self, request, exception):
        if isinstance(exception, ZeroDivisionError):
            raise ValueError("Value Error")
        else:
            base_response = ErrorHandler.response
            if hasattr(base_response, "__wrapped__"):
                base_response = base_response.__wrapped__

            return base_response(self, request, exception)

    def add(self, exception, handler, *args, **kwargs):
        base_add = ErrorHandler.add
        if hasattr(base_add, "__wrapped__"):
            base_add = base_add.__wrapped__
        base_add(self, exception, handler)


class CustomRouter(Router):
    def __init__(self):
        try:
            super().__init__(app=None)
        except TypeError:
            super().__init__()

    def add(self, *args, **kwargs):
        base_add = Router.add
        if hasattr(base_add, "__wrapped__"):
            base_add = base_add.__wrapped__
        return base_add.__get__(self, Router)(*args, **kwargs)

    def get(self, *args):
        base_get = Router.get
        if hasattr(base_get, "__wrapped__"):
            base_get = base_get.__wrapped__

        if len(args) == 1:
            path = args[0].path
        else:
            path = args[0]

        bound_get = base_get.__get__(self, Router)
        get_results = list(bound_get(*args))
        if path == "/server-error":
            from sanic.exceptions import ServerError

            raise ServerError("Server Error")
        return get_results


try:
    error_handler = CustomErrorHandler(fallback="text")
except TypeError:
    error_handler = CustomErrorHandler()

router = CustomRouter()
app = Sanic(name="test-app", error_handler=error_handler, router=router)
router.app = app
blueprint = Blueprint("test_bp")


@app.route("/")
async def index(request):
    return json({"hello": "world"})


@app.route("/error")
async def error(request):
    raise ValueError("Exception")


# see write_callback in confest.create_request_coroutine
@app.route("/write_response_error")
async def write_response_error(request):
    return json({"url": "write_response_error"})


@app.route("/404")
async def not_found(request):
    raise NotFound("Not found")


@app.route("/zero")
async def zero_division_error(request):
    1 / 0


@app.middleware("request")
async def request_middleware(request):
    return None


@blueprint.middleware("request")
async def blueprint_middleware(request):
    return None


# register the middleware a second time, testing that the `request_middleware`
# function is not getting double wrapped
app.register_middleware(request_middleware)


async def do_streaming(request):
    if stream is not None:
        # Old style response streaming
        async def streaming_fn(response):
            response.write("foo")
            response.write("bar")

        return stream(streaming_fn)
    else:
        # New style response streaming
        response = await request.respond(content_type="text/plain")
        await response.send("foo")
        await response.send("bar")
        await response.eof()


@app.route("/streaming")
async def streaming(request):
    return await do_streaming(request)


# Fake websocket endpoint to enable websockets on the server
@app.websocket("/socket")
async def socket(request, ws):
    assert False


@app.route("/custom-header/<header_key>/<header_value>")
async def custom_header(request, header_key, header_value):
    custom_headers = {header_key: header_value}
    return json({"hello": "world"}, headers=custom_headers)


@app.route("/server-error")
async def server_error(request):
    raise AssertionError("This handler should never be reached!")


class CustomExceptionSync(SanicException):
    pass


class CustomExceptionAsync(SanicException):
    pass


@app.exception(ServerError)
def handle_server_error(request, exception):
    pass


@app.exception(CustomExceptionSync)
def handle_custom_exception_sync(request, exception):
    raise SanicException("something went wrong")


@app.exception(CustomExceptionAsync)
async def handle_custom_exception_async(request, exception):
    raise SanicException("something went wrong")


@app.route("/sync-error")
async def sync_error(request):
    raise CustomExceptionSync("something went wrong")


@app.route("/async-error")
async def async_error(request):
    raise CustomExceptionAsync("something went wrong")


@blueprint.route("/blueprint")
async def blueprint_route(request):
    return await do_streaming(request)

app.blueprint(blueprint)
app.add_route(MethodView.as_view(), "/method_view")


if __name__ == "__main__":
    app.run(host="127.0.0.1", port=8000)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides functions to collect information about the operating
system, Python and hosting environment.

"""

import logging
import os
import platform
import sys

import newrelic
from newrelic.common.package_version_utils import get_package_version
from newrelic.common.system_info import (
    logical_processor_count,
    physical_processor_count,
    total_physical_memory,
)
from newrelic.core.config import global_settings
from newrelic.packages.isort import stdlibs as isort_stdlibs

try:
    import newrelic.core._thread_utilization
except ImportError:
    pass

_logger = logging.getLogger(__name__)


def environment_settings():
    """Returns an array of arrays of environment settings"""
    env = []

    # Agent information.

    env.append(("Agent Version", ".".join(map(str, newrelic.version_info))))

    # System information.

    env.append(("Arch", platform.machine()))
    env.append(("OS", platform.system()))
    env.append(("OS version", platform.release()))

    env.append(("Total Physical Memory (MB)", total_physical_memory()))
    env.append(("Logical Processors", logical_processor_count()))

    physical_processor_packages, physical_cores = physical_processor_count()

    # Report this attribute only if it has a valid value.

    if physical_processor_packages:
        env.append(("Physical Processor Packages", physical_processor_packages))

    # Report this attribute only if it has a valid value.

    if physical_cores:
        env.append(("Physical Cores", physical_cores))

    # Python information.

    env.append(("Python Program Name", sys.argv[0]))

    env.append(("Python Executable", sys.executable))

    env.append(("Python Home", os.environ.get("PYTHONHOME", "")))
    env.append(("Python Path", os.environ.get("PYTHONPATH", "")))

    env.append(("Python Prefix", sys.prefix))
    env.append(("Python Exec Prefix", sys.exec_prefix))

    env.append(("Python Runtime", ".".join(platform.python_version_tuple())))

    env.append(("Python Implementation", platform.python_implementation()))
    env.append(("Python Version", sys.version))

    env.append(("Python Platform", sys.platform))

    env.append(("Python Max Unicode", sys.maxunicode))

    # Extensions information.

    extensions = []

    if "newrelic.core._thread_utilization" in sys.modules:
        extensions.append("newrelic.core._thread_utilization")

    env.append(("Compiled Extensions", ", ".join(extensions)))

    # Dispatcher information.

    dispatcher = []

    # Find the first dispatcher module that's been loaded and report that as the dispatcher.
    # If possible, also report the dispatcher's version and any other environment information.
    if not dispatcher and "mod_wsgi" in sys.modules:
        mod_wsgi = sys.modules["mod_wsgi"]
        if hasattr(mod_wsgi, "process_group"):
            if mod_wsgi.process_group == "":
                dispatcher.append(("Dispatcher", "Apache/mod_wsgi (embedded)"))
            else:
                dispatcher.append(("Dispatcher", "Apache/mod_wsgi (daemon)"))
            env.append(("Apache/mod_wsgi Process Group", mod_wsgi.process_group))
        else:
            dispatcher.append(("Dispatcher", "Apache/mod_wsgi"))
        if hasattr(mod_wsgi, "version"):
            dispatcher.append(("Dispatcher Version", str(mod_wsgi.version)))
        if hasattr(mod_wsgi, "application_group"):
            env.append(("Apache/mod_wsgi Application Group", mod_wsgi.application_group))

    if not dispatcher and "uwsgi" in sys.modules:
        dispatcher.append(("Dispatcher", "uWSGI"))
        uwsgi = sys.modules["uwsgi"]
        if hasattr(uwsgi, "version"):
            dispatcher.append(("Dispatcher Version", uwsgi.version))

    if not dispatcher and "flup.server.fcgi" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/fastcgi (threaded)"))

    if not dispatcher and "flup.server.fcgi_fork" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/fastcgi (prefork)"))

    if not dispatcher and "flup.server.scgi" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/scgi (threaded)"))

    if not dispatcher and "flup.server.scgi_fork" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/scgi (prefork)"))

    if not dispatcher and "flup.server.ajp" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/ajp (threaded)"))

    if not dispatcher and "flup.server.ajp_fork" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/ajp (forking)"))

    if not dispatcher and "flup.server.cgi" in sys.modules:
        dispatcher.append(("Dispatcher", "flup/cgi"))

    if not dispatcher and "gunicorn" in sys.modules:
        if "gunicorn.workers.ggevent" in sys.modules:
            dispatcher.append(("Dispatcher", "gunicorn (gevent)"))
        elif "gunicorn.workers.geventlet" in sys.modules:
            dispatcher.append(("Dispatcher", "gunicorn (eventlet)"))
        elif "uvicorn.workers" in sys.modules:
            dispatcher.append(("Dispatcher", "gunicorn (uvicorn)"))
            uvicorn = sys.modules.get("uvicorn")
            if hasattr(uvicorn, "__version__"):
                dispatcher.append(("Worker Version", uvicorn.__version__))
        else:
            dispatcher.append(("Dispatcher", "gunicorn"))

        gunicorn = sys.modules["gunicorn"]
        if hasattr(gunicorn, "__version__"):
            dispatcher.append(("Dispatcher Version", gunicorn.__version__))

    if not dispatcher and "uvicorn" in sys.modules:
        dispatcher.append(("Dispatcher", "uvicorn"))
        uvicorn = sys.modules["uvicorn"]

        if hasattr(uvicorn, "__version__"):
            dispatcher.append(("Dispatcher Version", uvicorn.__version__))

    if not dispatcher and "hypercorn" in sys.modules:
        dispatcher.append(("Dispatcher", "hypercorn"))
        hypercorn = sys.modules["hypercorn"]

        if hasattr(hypercorn, "__version__"):
            dispatcher.append(("Dispatcher Version", hypercorn.__version__))
        else:
            try:
                dispatcher.append(("Dispatcher Version", get_package_version("hypercorn")))
            except Exception:
                pass

    if not dispatcher and "daphne" in sys.modules:
        dispatcher.append(("Dispatcher", "daphne"))
        daphne = sys.modules["daphne"]

        if hasattr(daphne, "__version__"):
            dispatcher.append(("Dispatcher Version", daphne.__version__))

    if not dispatcher and "tornado" in sys.modules:
        dispatcher.append(("Dispatcher", "tornado"))
        tornado = sys.modules["tornado"]
        if hasattr(tornado, "version_info"):
            dispatcher.append(("Dispatcher Version", str(tornado.version_info)))

    env.extend(dispatcher)

    # Module information.
    stdlib_builtin_module_names = _get_stdlib_builtin_module_names()

    plugins = []

    settings = global_settings()
    if settings and settings.package_reporting.enabled:
        # Using any iterable to create a snapshot of sys.modules can occassionally
        # fail in a rare case when modules are imported in parallel by different
        # threads.
        #
        # TL;DR: Do NOT use an iterable on the original sys.modules to generate the
        # list
        for name, module in sys.modules.copy().items():
            # Exclude lib.sub_paths as independent modules except for newrelic.hooks.
            nr_hook = name.startswith("newrelic.hooks.")
            if "." in name and not nr_hook or name.startswith("_"):
                continue

            # If the module isn't actually loaded (such as failed relative imports
            # in Python 2.7), the module will be None and should not be reported.
            try:
                if not module:
                    continue
            except Exception:
                # if the application uses generalimport to manage optional depedencies,
                # it's possible that generalimport.MissingOptionalDependency is raised.
                # In this case, we should not report the module as it is not actually loaded and
                # is not a runtime dependency of the application.
                #
                continue

            # Exclude standard library/built-in modules.
            if name in stdlib_builtin_module_names:
                continue

            try:
                version = get_package_version(name)
            except Exception:
                version = None

            # If it has no version it's likely not a real package so don't report it unless
            # it's a new relic hook.
            if version or nr_hook:
                plugins.append("%s (%s)" % (name, version))

    env.append(("Plugin List", plugins))

    return env


def _get_stdlib_builtin_module_names():
    builtins = set(sys.builtin_module_names)
    # Since sys.stdlib_module_names is not available in versions of python below 3.10,
    # use isort's hardcoded stdlibs instead.
    python_version = sys.version_info[0:2]
    if python_version < (3,):
        stdlibs = isort_stdlibs.py27.stdlib
    elif (3, 7) <= python_version < (3, 8):
        stdlibs = isort_stdlibs.py37.stdlib
    elif python_version < (3, 9):
        stdlibs = isort_stdlibs.py38.stdlib
    elif python_version < (3, 10):
        stdlibs = isort_stdlibs.py39.stdlib
    elif python_version >= (3, 10):
        stdlibs = sys.stdlib_module_names
    else:
        _logger.warn("Unsupported Python version. Unable to determine stdlibs.")
        return builtins
    return builtins | stdlibs



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import pytest
import pika

pika_version_info = tuple(int(num) for num in pika.__version__.split('.')[:2])

new_pika_xfail = pytest.mark.xfail(
        condition=pika_version_info[0] > 0, strict=True,
        reason='test fails if pika version is 1.x or greater')
new_pika_xfail_py37 = pytest.mark.xfail(
        condition=pika_version_info[0] > 0 and sys.version_info >= (3, 7),
        strict=True,
        reason='test fails if pika version is 1.x or greater')
new_pika_skip = pytest.mark.skipif(
        condition=pika_version_info[0] > 0,
        reason='test hangs if pika version is 1.x or greater')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest
import webtest
from testing_support.fixtures import (
    override_application_settings,
    validate_agent_attribute_types,
    validate_attributes,
    validate_attributes_complete,
)
from testing_support.sample_applications import fully_featured_app
from testing_support.validators.validate_custom_parameters import (
    validate_custom_parameters,
)

from newrelic.api.background_task import background_task
from newrelic.api.transaction import add_custom_attribute, add_custom_attributes
from newrelic.api.wsgi_application import wsgi_application
from newrelic.core.attribute import (
    _DESTINATIONS_WITH_EVENTS,
    MAX_64_BIT_INT,
    Attribute,
    CastingFailureException,
    sanitize,
    truncate,
)
from newrelic.packages import six

# Python 3 lacks longs

if sys.version_info >= (3, 0):
    long = int
try:
    from newrelic.core._thread_utilization import ThreadUtilization
except ImportError:
    ThreadUtilization = None


@wsgi_application()
def target_wsgi_application(environ, start_response):
    status = "200 OK"
    output = b"Hello World!"

    path = environ.get("PATH_INFO")
    if path == "/user_attribute":
        add_custom_attribute("test_key", "test_value")

    response_headers = [("Content-Type", "text/plain; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    return [output]


_required_intrinsics = ["trip_id", "totalTime"]
_forgone_intrinsics = []


@validate_attributes("intrinsic", _required_intrinsics, _forgone_intrinsics)
def test_intrinsics():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/")
    assert response.body == b"Hello World!"


_required_agent = [
    "request.method",
    "wsgi.output.seconds",
    "response.status",
    "request.headers.host",
    "request.headers.accept",
    "request.uri",
    "response.headers.contentType",
    "response.headers.contentLength",
]
if ThreadUtilization:
    _required_agent.append("thread.concurrency")
_forgone_agent = []


@validate_attributes("agent", _required_agent, _forgone_agent)
def test_agent():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/", extra_environ={"HTTP_ACCEPT": "*/*"})
    assert response.body == b"Hello World!"


_required_user = []
_forgone_user = ["test_key"]


@validate_attributes("user", _required_user, _forgone_user)
def test_user_default():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/")
    assert response.body == b"Hello World!"


_required_user = ["test_key"]
_forgone_user = []


@validate_attributes("user", _required_user, _forgone_user)
def test_user_add_attribute():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/user_attribute")
    assert response.body == b"Hello World!"


_settings_legacy_false = {"capture_params": False}
_required_request_legacy_false = []
_forgone_request_legacy_false = ["request.parameters.foo"]


@override_application_settings(_settings_legacy_false)
@validate_attributes("agent", _required_request_legacy_false, _forgone_request_legacy_false)
def test_capture_request_params_legacy_false():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/?foo=bar")
    assert response.body == b"Hello World!"


_settings_legacy_true = {"capture_params": True}
_required_request_legacy_true = ["request.parameters.foo"]
_forgone_request_legacy_true = []


@override_application_settings(_settings_legacy_true)
@validate_attributes("agent", _required_request_legacy_true, _forgone_request_legacy_true)
def test_capture_request_params_legacy_true():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/?foo=bar")
    assert response.body == b"Hello World!"


_required_request_default = ["request.parameters.foo"]
_forgone_request_default = []


@validate_attributes("agent", _required_request_default, _forgone_request_default)
def test_capture_request_params_default():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/?foo=bar")
    assert response.body == b"Hello World!"


_required_display_host_default = []
_forgone_display_host_default = ["host.displayName"]


@validate_attributes("agent", _required_display_host_default, _forgone_display_host_default)
def test_display_host_default():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/")
    assert response.body == b"Hello World!"


_settings_display_host_custom = {"process_host.display_name": "CUSTOM NAME"}

_display_name_attribute = Attribute(
    name="host.displayName", value="CUSTOM NAME", destinations=_DESTINATIONS_WITH_EVENTS
)
_required_display_host_custom = [_display_name_attribute]

_forgone_display_host_custom = []


@override_application_settings(_settings_display_host_custom)
@validate_attributes_complete("agent", _required_display_host_custom, _forgone_display_host_custom)
def test_display_host_custom():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/")
    assert response.body == b"Hello World!"


# Tests for truncate()


def test_truncate_string():
    s = "blahblah"
    result = truncate(s, maxsize=4)
    assert isinstance(result, six.string_types)
    assert result == "blah"


def test_truncate_bytes():
    b = b"foobar"
    result = truncate(b, maxsize=3)
    assert isinstance(result, six.binary_type)
    assert result == b"foo"


def test_truncate_unicode_snowman():
    # '\u2603' is 'SNOWMAN'
    # decode("unicode-escape") is used to get Py2 unicode
    u = "snow\u2603".decode("unicode-escape") if six.PY2 else "snow\u2603"
    assert u.encode("utf-8") == b"snow\xe2\x98\x83"
    result = truncate(u, maxsize=5)
    assert isinstance(result, six.text_type)
    assert result == "snow"


def test_truncate_combining_characters():
    # '\u0308' is 'COMBINING DIAERESIS' (AKA 'umlaut')
    # decode("unicode-escape") is used to get Py2 unicode
    u = "Zoe\u0308".decode("unicode-escape") if six.PY2 else "Zoe\u0308"
    assert u.encode("utf-8") == b"Zoe\xcc\x88"

    # truncate will chop off 'COMBINING DIAERESIS', which leaves
    # 'LATIN SMALL LETTER E' by itself.

    result = truncate(u, maxsize=3)
    assert isinstance(result, six.text_type)
    assert result == "Zoe"


def test_truncate_empty_string():
    s = ""
    result = truncate(s, maxsize=4)
    assert isinstance(result, six.string_types)
    assert result == ""


def test_truncate_empty_bytes():
    b = b""
    result = truncate(b, maxsize=3)
    assert isinstance(result, six.binary_type)
    assert result == b""


def test_truncate_empty_unicode():
    # decode("unicode-escape") is used to get Py2 unicode
    u = "".decode("unicode-escape") if six.PY2 else ""
    result = truncate(u, maxsize=5)
    assert isinstance(result, six.text_type)
    assert result == ""


# Tests for limits on user attributes

TOO_LONG = "*" * 256
TRUNCATED = "*" * 255


_required_custom_params = [("key", "value")]
_forgone_custom_params = []


@validate_custom_parameters(_required_custom_params, _forgone_custom_params)
@background_task()
def test_custom_param_ok():
    result = add_custom_attribute("key", "value")
    assert result


@validate_custom_parameters(_required_custom_params, _forgone_custom_params)
@background_task()
def test_custom_params_ok():
    result = add_custom_attributes([("key", "value")])
    assert result


_required_custom_params_long_key = []
_forgone_custom_params_long_key = [(TOO_LONG, "value")]


@validate_custom_parameters(_required_custom_params_long_key, _forgone_custom_params_long_key)
@background_task()
def test_custom_param_key_too_long():
    result = add_custom_attribute(TOO_LONG, "value")
    assert not result


@validate_custom_parameters(_required_custom_params_long_key, _forgone_custom_params_long_key)
@background_task()
def test_custom_params_key_too_long():
    result = add_custom_attributes([(TOO_LONG, "value")])
    assert not result


_required_custom_params_long_value = [("key", TRUNCATED)]
_forgone_custom_params_long_value = []


@validate_custom_parameters(_required_custom_params_long_value, _forgone_custom_params_long_value)
@background_task()
def test_custom_param_value_too_long():
    result = add_custom_attribute("key", TOO_LONG)
    assert result


@validate_custom_parameters(_required_custom_params_long_value, _forgone_custom_params_long_value)
@background_task()
def test_custom_params_value_too_long():
    result = add_custom_attributes([("key", TOO_LONG)])
    assert result


_required_custom_params_too_many = [("key-127", "value")]
_forgone_custom_params_too_many = [("key-128", "value")]


@validate_custom_parameters(_required_custom_params_too_many, _forgone_custom_params_too_many)
@background_task()
def test_custom_param_too_many():
    for i in range(129):
        result = add_custom_attribute("key-%02d" % i, "value")
        if i < 128:
            assert result
        else:
            assert not result  # Last one fails


@validate_custom_parameters(_required_custom_params_too_many, _forgone_custom_params_too_many)
@background_task()
def test_custom_params_too_many():
    item_list = [("key-%02d" % i, "value") for i in range(129)]
    result = add_custom_attributes(item_list)
    assert not result


_required_custom_params_name_not_string = []
_forgone_custom_params_name_not_string = [(1, "value")]


@validate_custom_parameters(_required_custom_params_name_not_string, _forgone_custom_params_name_not_string)
@background_task()
def test_custom_param_name_not_string():
    result = add_custom_attribute(1, "value")
    assert not result


@validate_custom_parameters(_required_custom_params_name_not_string, _forgone_custom_params_name_not_string)
@background_task()
def test_custom_params_name_not_string():
    result = add_custom_attributes([(1, "value")])
    assert not result


TOO_BIG = MAX_64_BIT_INT + 1

_required_custom_params_int_too_big = []
_forgone_custom_params_int_too_big = [("key", TOO_BIG)]


@validate_custom_parameters(_required_custom_params_int_too_big, _forgone_custom_params_int_too_big)
@background_task()
def test_custom_param_int_too_big():
    result = add_custom_attribute("key", TOO_BIG)
    assert not result


@validate_custom_parameters(_required_custom_params_int_too_big, _forgone_custom_params_int_too_big)
@background_task()
def test_custom_params_int_too_big():
    result = add_custom_attributes([("key", TOO_BIG)])
    assert not result


OK_KEY = "*" * (255 - len("request.parameters."))
OK_REQUEST_PARAM = "request.parameters." + OK_KEY
TOO_LONG_KEY = "*" * (256 - len("request.parameters."))
TOO_LONG_REQUEST_PARAM = "request.parameters." + TOO_LONG_KEY

assert len(OK_REQUEST_PARAM) == 255
assert len(TOO_LONG_REQUEST_PARAM) == 256

_required_request_key_ok = [OK_REQUEST_PARAM]
_forgone_request_key_ok = []


@validate_attributes("agent", _required_request_key_ok, _forgone_request_key_ok)
def test_capture_request_params_key_ok():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/?%s=bar" % OK_KEY)
    assert response.body == b"Hello World!"


_required_request_key_too_long = []
_forgone_request_key_too_long = [TOO_LONG_REQUEST_PARAM]


@validate_attributes("agent", _required_request_key_too_long, _forgone_request_key_too_long)
def test_capture_request_params_key_too_long():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/?%s=bar" % TOO_LONG_KEY)
    assert response.body == b"Hello World!"


_required_request_value_too_long = ["request.parameters.foo"]
_forgone_request_value_too_long = []


@validate_attributes("agent", _required_request_value_too_long, _forgone_request_value_too_long)
def test_capture_request_params_value_too_long():
    target_application = webtest.TestApp(target_wsgi_application)
    response = target_application.get("/?foo=%s" % TOO_LONG)
    assert response.body == b"Hello World!"


# Test attribute types are according to Agent-Attributes spec.

fully_featured_application = webtest.TestApp(fully_featured_app)

# Types are only defined in the spec for agent attributes, not intrinsics.

agent_attributes = {
    "request.headers.accept": six.string_types,
    "request.headers.contentLength": int,
    "request.headers.contentType": six.string_types,
    "request.headers.host": six.string_types,
    "request.headers.referer": six.string_types,
    "request.headers.userAgent": six.string_types,
    "request.method": six.string_types,
    "request.parameters.test": six.string_types,
    "response.headers.contentLength": int,
    "response.headers.contentType": six.string_types,
    "response.status": six.string_types,
}


@validate_agent_attribute_types(agent_attributes)
def test_agent_attribute_types():
    test_environ = {
        "CONTENT_TYPE": "HTML",
        "CONTENT_LENGTH": "100",
        "HTTP_USER_AGENT": "Firefox",
        "HTTP_REFERER": "somewhere",
        "HTTP_ACCEPT": "everything",
    }
    fully_featured_application.get("/?test=val", extra_environ=test_environ)


# Test sanitize()


def test_sanitize_string():
    s = "foo"
    assert sanitize(s) == s


def test_sanitize_bytes():
    b = b"bytes"
    assert sanitize(b) == b


def test_sanitize_unicode():
    u = "SMILING FACE: \u263a"
    assert sanitize(u) == u


def test_sanitize_bool():
    assert sanitize(True) is True


def test_sanitize_float():
    assert sanitize(1.11) == 1.11


def test_sanitize_int():
    assert sanitize(9876) == 9876


def test_sanitize_long():
    long_int = long(123456)
    assert sanitize(long_int) == long_int


def test_sanitize_dict():
    d = {1: "foo"}
    assert sanitize(d) == "{1: 'foo'}"


def test_sanitize_list():
    list_var = [1, 2, 3, 4]
    assert sanitize(list_var) == "[1, 2, 3, 4]"


def test_sanitize_tuple():
    t = ("one", "two", "three")
    assert sanitize(t) == "('one', 'two', 'three')"


class Foo(object):
    pass


def test_sanitize_object():
    f = Foo()
    assert sanitize(f) == str(f)


class TypeErrorString(object):
    def __str__(self):
        return 42


def test_str_raises_type_error():
    with pytest.raises(CastingFailureException):
        sanitize(TypeErrorString())


class AttributeErrorString(object):
    def __str__(self):
        raise AttributeError()


def test_str_raises_attribute_error():
    with pytest.raises(CastingFailureException):
        sanitize(AttributeErrorString())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from testing_support.fixtures import validate_tt_parenting
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

expected_parenting = (
    "TransactionNode",
    [
        (
            "FunctionNode",
            [
                ("ExternalTrace", []),
                ("ExternalTrace", []),
            ],
        ),
    ],
)


@validate_tt_parenting(expected_parenting)
@validate_transaction_metrics("_target_application:multi_fetch_handler", rollup_metrics=[("External/all", 2)])
def test_multiple_requests_within_transaction(local_server_info, aiohttp_app):
    async def fetch():
        resp = await aiohttp_app.client.request("GET", "/multi_fetch", params={"url": local_server_info.url})
        assert resp.status == 200

    aiohttp_app.loop.run_until_complete(fetch())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version
from newrelic.packages import six

SKLEARN_VERSION = tuple(map(int, get_package_version("sklearn").split(".")))
SCIPY_VERSION = tuple(map(int, get_package_version("scipy").split(".")))


@pytest.mark.parametrize(
    "linear_model_name",
    [
        "ARDRegression",
        "BayesianRidge",
        "ElasticNet",
        "ElasticNetCV",
        "HuberRegressor",
        "Lars",
        "LarsCV",
        "Lasso",
        "LassoCV",
        "LassoLars",
        "LassoLarsCV",
        "LassoLarsIC",
        "LinearRegression",
        "LogisticRegression",
        "LogisticRegressionCV",
        "MultiTaskElasticNet",
        "MultiTaskElasticNetCV",
        "MultiTaskLasso",
        "MultiTaskLassoCV",
        "OrthogonalMatchingPursuit",
        "OrthogonalMatchingPursuitCV",
        "PassiveAggressiveClassifier",
        "PassiveAggressiveRegressor",
        "Perceptron",
        "Ridge",
        "RidgeCV",
        "RidgeClassifier",
        "RidgeClassifierCV",
        "TheilSenRegressor",
        "RANSACRegressor",
    ],
)
def test_model_methods_wrapped_in_function_trace(linear_model_name, run_linear_model):
    expected_scoped_metrics = {
        "ARDRegression": [
            ("Function/MLModel/Sklearn/Named/ARDRegression.fit", 1),
            ("Function/MLModel/Sklearn/Named/ARDRegression.predict", 2),
            ("Function/MLModel/Sklearn/Named/ARDRegression.score", 1),
        ],
        "BayesianRidge": [
            ("Function/MLModel/Sklearn/Named/BayesianRidge.fit", 1),
            ("Function/MLModel/Sklearn/Named/BayesianRidge.predict", 2),
            ("Function/MLModel/Sklearn/Named/BayesianRidge.score", 1),
        ],
        "ElasticNet": [
            ("Function/MLModel/Sklearn/Named/ElasticNet.fit", 1),
            ("Function/MLModel/Sklearn/Named/ElasticNet.predict", 2),
            ("Function/MLModel/Sklearn/Named/ElasticNet.score", 1),
        ],
        "ElasticNetCV": [
            ("Function/MLModel/Sklearn/Named/ElasticNetCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/ElasticNetCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/ElasticNetCV.score", 1),
        ],
        "HuberRegressor": [
            ("Function/MLModel/Sklearn/Named/HuberRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/HuberRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/HuberRegressor.score", 1),
        ],
        "Lars": [
            ("Function/MLModel/Sklearn/Named/Lars.fit", 1),
            ("Function/MLModel/Sklearn/Named/Lars.predict", 2),
            ("Function/MLModel/Sklearn/Named/Lars.score", 1),
        ],
        "LarsCV": [
            ("Function/MLModel/Sklearn/Named/LarsCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/LarsCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/LarsCV.score", 1),
        ],
        "Lasso": [
            ("Function/MLModel/Sklearn/Named/Lasso.fit", 1),
            ("Function/MLModel/Sklearn/Named/Lasso.predict", 2),
            ("Function/MLModel/Sklearn/Named/Lasso.score", 1),
        ],
        "LassoCV": [
            ("Function/MLModel/Sklearn/Named/LassoCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/LassoCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/LassoCV.score", 1),
        ],
        "LassoLars": [
            ("Function/MLModel/Sklearn/Named/LassoLars.fit", 1),
            ("Function/MLModel/Sklearn/Named/LassoLars.predict", 2),
            ("Function/MLModel/Sklearn/Named/LassoLars.score", 1),
        ],
        "LassoLarsCV": [
            ("Function/MLModel/Sklearn/Named/LassoLarsCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/LassoLarsCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/LassoLarsCV.score", 1),
        ],
        "LassoLarsIC": [
            ("Function/MLModel/Sklearn/Named/LassoLarsIC.fit", 1),
            ("Function/MLModel/Sklearn/Named/LassoLarsIC.predict", 2),
            ("Function/MLModel/Sklearn/Named/LassoLarsIC.score", 1),
        ],
        "LinearRegression": [
            ("Function/MLModel/Sklearn/Named/LinearRegression.fit", 1),
            ("Function/MLModel/Sklearn/Named/LinearRegression.predict", 2),
            ("Function/MLModel/Sklearn/Named/LinearRegression.score", 1),
        ],
        "LogisticRegression": [
            ("Function/MLModel/Sklearn/Named/LogisticRegression.fit", 1),
            ("Function/MLModel/Sklearn/Named/LogisticRegression.predict", 2),
            ("Function/MLModel/Sklearn/Named/LogisticRegression.score", 1),
        ],
        "LogisticRegressionCV": [
            ("Function/MLModel/Sklearn/Named/LogisticRegressionCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/LogisticRegressionCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/LogisticRegressionCV.score", 1),
        ],
        "MultiTaskElasticNet": [
            ("Function/MLModel/Sklearn/Named/MultiTaskElasticNet.fit", 1),
            ("Function/MLModel/Sklearn/Named/MultiTaskElasticNet.predict", 2),
            ("Function/MLModel/Sklearn/Named/MultiTaskElasticNet.score", 1),
        ],
        "MultiTaskElasticNetCV": [
            ("Function/MLModel/Sklearn/Named/MultiTaskElasticNetCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/MultiTaskElasticNetCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/MultiTaskElasticNetCV.score", 1),
        ],
        "MultiTaskLasso": [
            ("Function/MLModel/Sklearn/Named/MultiTaskLasso.fit", 1),
            ("Function/MLModel/Sklearn/Named/MultiTaskLasso.predict", 2),
            ("Function/MLModel/Sklearn/Named/MultiTaskLasso.score", 1),
        ],
        "MultiTaskLassoCV": [
            ("Function/MLModel/Sklearn/Named/MultiTaskLassoCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/MultiTaskLassoCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/MultiTaskLassoCV.score", 1),
        ],
        "OrthogonalMatchingPursuit": [
            ("Function/MLModel/Sklearn/Named/OrthogonalMatchingPursuit.fit", 1),
            ("Function/MLModel/Sklearn/Named/OrthogonalMatchingPursuit.predict", 2),
            ("Function/MLModel/Sklearn/Named/OrthogonalMatchingPursuit.score", 1),
        ],
        "OrthogonalMatchingPursuitCV": [
            ("Function/MLModel/Sklearn/Named/OrthogonalMatchingPursuitCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/OrthogonalMatchingPursuitCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/OrthogonalMatchingPursuitCV.score", 1),
        ],
        "PassiveAggressiveClassifier": [
            ("Function/MLModel/Sklearn/Named/PassiveAggressiveClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/PassiveAggressiveClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/PassiveAggressiveClassifier.score", 1),
        ],
        "PassiveAggressiveRegressor": [
            ("Function/MLModel/Sklearn/Named/PassiveAggressiveRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/PassiveAggressiveRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/PassiveAggressiveRegressor.score", 1),
        ],
        "Perceptron": [
            ("Function/MLModel/Sklearn/Named/Perceptron.fit", 1),
            ("Function/MLModel/Sklearn/Named/Perceptron.predict", 2),
            ("Function/MLModel/Sklearn/Named/Perceptron.score", 1),
        ],
        "Ridge": [
            ("Function/MLModel/Sklearn/Named/Ridge.fit", 1),
            ("Function/MLModel/Sklearn/Named/Ridge.predict", 2),
            ("Function/MLModel/Sklearn/Named/Ridge.score", 1),
        ],
        "RidgeCV": [
            ("Function/MLModel/Sklearn/Named/RidgeCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/RidgeCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/RidgeCV.score", 1),
        ],
        "RidgeClassifier": [
            ("Function/MLModel/Sklearn/Named/RidgeClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/RidgeClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/RidgeClassifier.score", 1),
        ],
        "RidgeClassifierCV": [
            ("Function/MLModel/Sklearn/Named/RidgeClassifierCV.fit", 1),
            ("Function/MLModel/Sklearn/Named/RidgeClassifierCV.predict", 2),
            ("Function/MLModel/Sklearn/Named/RidgeClassifierCV.score", 1),
        ],
        "TheilSenRegressor": [
            ("Function/MLModel/Sklearn/Named/TheilSenRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/TheilSenRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/TheilSenRegressor.score", 1),
        ],
        "RANSACRegressor": [
            ("Function/MLModel/Sklearn/Named/RANSACRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/RANSACRegressor.predict", 1),
            ("Function/MLModel/Sklearn/Named/RANSACRegressor.score", 1),
        ],
    }
    expected_transaction_name = "test_linear_models:_test"
    if six.PY3:
        expected_transaction_name = "test_linear_models:test_model_methods_wrapped_in_function_trace.<locals>._test"

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[linear_model_name],
        rollup_metrics=expected_scoped_metrics[linear_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_linear_model(linear_model_name)

    _test()


@pytest.mark.skipif(SKLEARN_VERSION < (1, 1, 0), reason="Requires sklearn >= v1.1")
@pytest.mark.parametrize(
    "linear_model_name",
    [
        "PoissonRegressor",
        "GammaRegressor",
        "TweedieRegressor",
        "QuantileRegressor",
        "SGDClassifier",
        "SGDRegressor",
        "SGDOneClassSVM",
    ],
)
def test_above_v1_1_model_methods_wrapped_in_function_trace(linear_model_name, run_linear_model):
    expected_scoped_metrics = {
        "PoissonRegressor": [
            ("Function/MLModel/Sklearn/Named/PoissonRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/PoissonRegressor.predict", 1),
            ("Function/MLModel/Sklearn/Named/PoissonRegressor.score", 1),
        ],
        "GammaRegressor": [
            ("Function/MLModel/Sklearn/Named/GammaRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/GammaRegressor.predict", 1),
            ("Function/MLModel/Sklearn/Named/GammaRegressor.score", 1),
        ],
        "TweedieRegressor": [
            ("Function/MLModel/Sklearn/Named/TweedieRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/TweedieRegressor.predict", 1),
            ("Function/MLModel/Sklearn/Named/TweedieRegressor.score", 1),
        ],
        "QuantileRegressor": [
            ("Function/MLModel/Sklearn/Named/QuantileRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/QuantileRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/QuantileRegressor.score", 1),
        ],
        "SGDClassifier": [
            ("Function/MLModel/Sklearn/Named/SGDClassifier.fit", 1),
            ("Function/MLModel/Sklearn/Named/SGDClassifier.predict", 2),
            ("Function/MLModel/Sklearn/Named/SGDClassifier.score", 1),
        ],
        "SGDRegressor": [
            ("Function/MLModel/Sklearn/Named/SGDRegressor.fit", 1),
            ("Function/MLModel/Sklearn/Named/SGDRegressor.predict", 2),
            ("Function/MLModel/Sklearn/Named/SGDRegressor.score", 1),
        ],
        "SGDOneClassSVM": [
            ("Function/MLModel/Sklearn/Named/SGDOneClassSVM.fit", 1),
            ("Function/MLModel/Sklearn/Named/SGDOneClassSVM.predict", 1),
        ],
    }
    expected_transaction_name = "test_linear_models:_test"
    if six.PY3:
        expected_transaction_name = (
            "test_linear_models:test_above_v1_1_model_methods_wrapped_in_function_trace.<locals>._test"
        )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[linear_model_name],
        rollup_metrics=expected_scoped_metrics[linear_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_linear_model(linear_model_name)

    _test()


@pytest.fixture
def run_linear_model():
    def _run(linear_model_name):
        import sklearn.linear_model
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        X, y = load_iris(return_X_y=True)
        x_train, x_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

        if linear_model_name == "GammaRegressor":
            x_train = [[1, 2], [2, 3], [3, 4], [4, 3]]
            y_train = [19, 26, 33, 30]
            x_test = [[1, 2], [2, 3], [3, 4], [4, 3]]
            y_test = [19, 26, 33, 30]
        elif linear_model_name in [
            "MultiTaskElasticNet",
            "MultiTaskElasticNetCV",
            "MultiTaskLasso",
            "MultiTaskLassoCV",
        ]:
            y_train = x_train
            y_test = x_test

        clf = getattr(sklearn.linear_model, linear_model_name)()

        if linear_model_name == "QuantileRegressor" and SCIPY_VERSION > (1, 11, 0):
            # Silence warnings and errors related to solver change
            clf.solver = "highs"

        model = clf.fit(x_train, y_train)
        model.predict(x_test)

        if hasattr(model, "score"):
            model.score(x_test, y_test)

        return model

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import requests

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_span_events import (
        validate_span_events)
from testing_support.mock_external_http_server import MockExternalHTTPServer
from newrelic.api.background_task import background_task
from newrelic.api.transaction import current_transaction


@pytest.fixture(scope='module', autouse=True)
def server():
    with MockExternalHTTPServer() as _server:
        yield _server


@pytest.mark.parametrize('path', ('', '/foo', '/' + 'a' * 256))
def test_span_events(server, path):
    _settings = {
        'distributed_tracing.enabled': True,
        'span_events.enabled': True,
    }

    uri = 'http://localhost:%d' % server.port
    if path:
        uri += path

    # Validate long uri gets truncated
    expected_uri = uri[:255]

    exact_intrinsics = {
        'name': 'External/localhost:%d/requests/' % server.port,
        'type': 'Span',
        'sampled': True,
        'priority': 0.5,

        'category': 'http',
        'span.kind': 'client',
        'component': 'requests',
    }
    exact_agents = {
        'http.url': expected_uri,
    }
    expected_intrinsics = ('timestamp', 'duration', 'transactionId')

    @override_application_settings(_settings)
    @validate_span_events(
            count=1,
            exact_intrinsics=exact_intrinsics,
            exact_agents=exact_agents,
            expected_intrinsics=expected_intrinsics)
    @background_task(name='test_span_events')
    def _test():
        txn = current_transaction()
        txn._priority = 0.5
        txn._sampled = True

        response = requests.get(uri)
        response.raise_for_status()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import sys

import pytest
from testing_support.fixtures import (
    capture_transaction_metrics,
    override_generic_settings,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)

from newrelic.api.background_task import background_task
from newrelic.api.function_trace import function_trace
from newrelic.api.message_transaction import message_transaction
from newrelic.api.transaction import current_transaction
from newrelic.api.web_transaction import web_transaction
from newrelic.core.config import global_settings

if sys.version_info >= (3, 5):
    from _test_async_coroutine_transaction import native_coroutine_test
else:
    native_coroutine_test = None

settings = global_settings()


def coroutine_test(event_loop, transaction, nr_enabled=True, does_hang=False, call_exit=False, runtime_error=False):
    @transaction
    async def task():
        txn = current_transaction()

        if not nr_enabled:
            assert txn is None
        else:
            assert txn._loop_time == 0.0

        if call_exit:
            txn.__exit__(None, None, None)
        else:
            assert current_transaction() is txn

        try:
            if does_hang:
                await loop.create_future()  # noqa
            else:
                await asyncio.sleep(0.0)
                if nr_enabled and txn.enabled:
                    # Validate loop time is recorded after suspend
                    assert txn._loop_time > 0.0
        except GeneratorExit:
            if runtime_error:
                await asyncio.sleep(0.0)

    return task


test_matrix = [coroutine_test]
if native_coroutine_test:
    test_matrix.append(native_coroutine_test)


@pytest.mark.parametrize("num_coroutines", (2,))
@pytest.mark.parametrize("create_test_task", test_matrix)
@pytest.mark.parametrize(
    "transaction,metric",
    [
        (background_task(name="test"), "OtherTransaction/Function/test"),
        (
            message_transaction("lib", "dest_type", "dest_name"),
            "OtherTransaction/Message/lib/dest_type/Named/dest_name",
        ),
    ],
)
@pytest.mark.parametrize(
    "nr_enabled,call_exit",
    (
        (False, False),
        (True, False),
        (True, True),
    ),
)
def test_async_coroutine_send(event_loop, num_coroutines, create_test_task, transaction, metric, call_exit, nr_enabled):
    metrics = []

    tasks = [
        create_test_task(event_loop, transaction, nr_enabled=nr_enabled, call_exit=call_exit)
        for _ in range(num_coroutines)
    ]

    @override_generic_settings(settings, {"enabled": nr_enabled})
    @capture_transaction_metrics(metrics)
    def _test_async_coroutine_send():
        driver = asyncio.gather(*[t() for t in tasks])
        event_loop.run_until_complete(driver)

    _test_async_coroutine_send()

    if nr_enabled:
        assert metrics.count((metric, "")) == num_coroutines, metrics
    else:
        assert not metrics, metrics


@pytest.mark.parametrize("num_coroutines", (2,))
@pytest.mark.parametrize("create_test_task", test_matrix)
@pytest.mark.parametrize(
    "transaction,metric",
    [
        (background_task(name="test"), "OtherTransaction/Function/test"),
        (
            message_transaction("lib", "dest_type", "dest_name"),
            "OtherTransaction/Message/lib/dest_type/Named/dest_name",
        ),
    ],
)
def test_async_coroutine_send_disabled(event_loop, num_coroutines, create_test_task, transaction, metric):
    metrics = []

    tasks = [create_test_task(event_loop, transaction, call_exit=True) for _ in range(num_coroutines)]

    @capture_transaction_metrics(metrics)
    def _test_async_coroutine_send():
        driver = asyncio.gather(*[t() for t in tasks])
        event_loop.run_until_complete(driver)

    _test_async_coroutine_send()

    assert metrics.count((metric, "")) == num_coroutines, metrics


@pytest.mark.parametrize("num_coroutines", (2,))
@pytest.mark.parametrize("create_test_task", test_matrix)
@pytest.mark.parametrize(
    "transaction,metric",
    [
        (background_task(name="test"), "OtherTransaction/Function/test"),
        (
            message_transaction("lib", "dest_type", "dest_name"),
            "OtherTransaction/Message/lib/dest_type/Named/dest_name",
        ),
    ],
)
@validate_transaction_errors([])
def test_async_coroutine_throw_cancel(event_loop, num_coroutines, create_test_task, transaction, metric):
    metrics = []

    tasks = [create_test_task(event_loop, transaction) for _ in range(num_coroutines)]

    async def task_c():
        futures = [asyncio.ensure_future(t()) for t in tasks]

        await asyncio.sleep(0.0)

        [f.cancel() for f in futures]

    @capture_transaction_metrics(metrics)
    def _test_async_coroutine_throw_cancel():
        event_loop.run_until_complete(task_c())

    _test_async_coroutine_throw_cancel()

    assert metrics.count((metric, "")) == num_coroutines, metrics


@pytest.mark.parametrize("num_coroutines", (2,))
@pytest.mark.parametrize("create_test_task", test_matrix)
@pytest.mark.parametrize(
    "transaction,metric",
    [
        (background_task(name="test"), "OtherTransaction/Function/test"),
        (
            message_transaction("lib", "dest_type", "dest_name"),
            "OtherTransaction/Message/lib/dest_type/Named/dest_name",
        ),
    ],
)
@validate_transaction_errors(["builtins:ValueError"])
def test_async_coroutine_throw_error(event_loop, num_coroutines, create_test_task, transaction, metric):
    metrics = []

    tasks = [create_test_task(event_loop, transaction) for _ in range(num_coroutines)]

    async def task_c():
        coros = [t() for t in tasks]

        for coro in coros:
            with pytest.raises(ValueError):
                coro.throw(ValueError)

    @capture_transaction_metrics(metrics)
    def _test_async_coroutine_throw_error():
        event_loop.run_until_complete(task_c())

    _test_async_coroutine_throw_error()

    assert metrics.count((metric, "")) == num_coroutines, metrics
    assert metrics.count(("Errors/" + metric, "")) == num_coroutines, metrics
    assert metrics.count(("Errors/all", "")) == num_coroutines, metrics


@pytest.mark.parametrize("num_coroutines", (1,))
@pytest.mark.parametrize("create_test_task", test_matrix)
@pytest.mark.parametrize(
    "transaction,metric",
    [
        (background_task(name="test"), "OtherTransaction/Function/test"),
        (
            message_transaction("lib", "dest_type", "dest_name"),
            "OtherTransaction/Message/lib/dest_type/Named/dest_name",
        ),
    ],
)
@pytest.mark.parametrize("start_coroutines", (False, True))
def test_async_coroutine_close(event_loop, num_coroutines, create_test_task, transaction, metric, start_coroutines):
    metrics = []

    tasks = [create_test_task(event_loop, transaction) for _ in range(num_coroutines)]

    async def task_c():
        coros = [t() for t in tasks]

        if start_coroutines:
            [asyncio.ensure_future(coro) for coro in coros]

            await asyncio.sleep(0.0)

        [coro.close() for coro in coros]

    @capture_transaction_metrics(metrics)
    def _test_async_coroutine_close():
        event_loop.run_until_complete(task_c())

    _test_async_coroutine_close()

    if start_coroutines:
        assert metrics.count((metric, "")) == num_coroutines, metrics
    else:
        assert not metrics


@pytest.mark.parametrize("num_coroutines", (1,))
@pytest.mark.parametrize("create_test_task", test_matrix)
@pytest.mark.parametrize(
    "transaction,metric",
    [
        (background_task(name="test"), "OtherTransaction/Function/test"),
        (
            message_transaction("lib", "dest_type", "dest_name"),
            "OtherTransaction/Message/lib/dest_type/Named/dest_name",
        ),
    ],
)
@validate_transaction_errors(["builtins:RuntimeError"])
def test_async_coroutine_close_raises_error(event_loop, num_coroutines, create_test_task, transaction, metric):
    metrics = []

    tasks = [create_test_task(event_loop, transaction, runtime_error=True) for _ in range(num_coroutines)]

    async def task_c():
        coros = [t() for t in tasks]

        [c.send(None) for c in coros]

        await asyncio.sleep(0.0)

        for coro in coros:
            with pytest.raises(RuntimeError):
                coro.close()

    @capture_transaction_metrics(metrics)
    def _test_async_coroutine_close_raises_error():
        event_loop.run_until_complete(task_c())

    _test_async_coroutine_close_raises_error()

    assert metrics.count((metric, "")) == num_coroutines, metrics
    assert metrics.count(("Errors/all", "")) == num_coroutines, metrics


@pytest.mark.parametrize(
    "transaction,metric,arguments",
    [
        (web_transaction, "Apdex/Function/%s", lambda name: ([], {"name": name})),
        (
            message_transaction,
            "OtherTransaction/Message/lib/dest_type/Named/%s",
            lambda name: (["lib", "dest_type", name], {}),
        ),
        (background_task, "OtherTransaction/Function/%s", lambda name: ([], {"name": name})),
    ],
)
def test_deferred_async_background_task(event_loop, transaction, metric, arguments):
    deferred_metric = (metric % "deferred", "")

    args, kwargs = arguments("deferred")

    @transaction(*args, **kwargs)
    async def child_task():
        await asyncio.sleep(0)

    main_metric = (metric % "main", "")

    args, kwargs = arguments("main")

    @transaction(*args, **kwargs)
    async def parent_task():
        await asyncio.sleep(0)
        return event_loop.create_task(child_task())

    async def test_runner():
        child = await parent_task()
        await child

    metrics = []

    @capture_transaction_metrics(metrics)
    def _test():
        event_loop.run_until_complete(test_runner())

    _test()

    assert main_metric in metrics
    assert deferred_metric in metrics


@pytest.mark.parametrize(
    "transaction,metric,arguments",
    [
        (web_transaction, "Apdex/Function/%s", lambda name: ([], {"name": name})),
        (
            message_transaction,
            "OtherTransaction/Message/lib/dest_type/Named/%s",
            lambda name: (["lib", "dest_type", name], {}),
        ),
        (background_task, "OtherTransaction/Function/%s", lambda name: ([], {"name": name})),
    ],
)
def test_child_transaction_when_parent_is_running(event_loop, transaction, metric, arguments):
    deferred_metric = (metric % "deferred", "")

    args, kwargs = arguments("deferred")

    @transaction(*args, **kwargs)
    async def child_task():
        await asyncio.sleep(0)

    main_metric = (metric % "main", "")

    args, kwargs = arguments("main")

    @transaction(*args, **kwargs)
    async def parent_task():
        await event_loop.create_task(child_task())

    metrics = []

    @capture_transaction_metrics(metrics)
    def _test():
        event_loop.run_until_complete(parent_task())

    _test()

    assert main_metric in metrics
    assert deferred_metric in metrics


@pytest.mark.parametrize(
    "transaction,metric,arguments",
    [
        (web_transaction, "Apdex/Function/%s", lambda name: ([], {"name": name})),
        (
            message_transaction,
            "OtherTransaction/Message/lib/dest_type/Named/%s",
            lambda name: (["lib", "dest_type", name], {}),
        ),
        (background_task, "OtherTransaction/Function/%s", lambda name: ([], {"name": name})),
    ],
)
def test_nested_coroutine_inside_sync(event_loop, transaction, metric, arguments):
    child_metric = (metric % "child", "")

    args, kwargs = arguments("child")

    @transaction(*args, **kwargs)
    async def child_task():
        await asyncio.sleep(0)

    main_metric = (metric % "main", "")
    args, kwargs = arguments("main")

    metrics = []

    @capture_transaction_metrics(metrics)
    @transaction(*args, **kwargs)
    def parent():
        event_loop.run_until_complete(child_task())

    parent()

    assert main_metric in metrics
    assert child_metric not in metrics


@pytest.mark.parametrize(
    "transaction,metric,arguments",
    [
        (web_transaction, "Apdex/Function/%s", lambda name: ([], {"name": name})),
        (
            message_transaction,
            "OtherTransaction/Message/lib/dest_type/Named/%s",
            lambda name: (["lib", "dest_type", name], {}),
        ),
        (background_task, "OtherTransaction/Function/%s", lambda name: ([], {"name": name})),
    ],
)
def test_nested_coroutine_task_already_active(event_loop, transaction, metric, arguments):
    deferred_metric = (metric % "deferred", "")

    args, kwargs = arguments("deferred")

    @transaction(*args, **kwargs)
    async def child_task():
        await asyncio.sleep(0)

    @function_trace()
    async def child_trace():
        await child_task()

    main_metric = (metric % "main", "")

    args, kwargs = arguments("main")

    @transaction(*args, **kwargs)
    async def parent_task():
        await event_loop.create_task(child_trace())

    metrics = []

    @capture_transaction_metrics(metrics)
    def _test():
        event_loop.run_until_complete(parent_task())

    _test()

    assert main_metric in metrics
    assert deferred_metric not in metrics



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_dimensional_metric_payload import (
    validate_dimensional_metric_payload,
)
from testing_support.validators.validate_dimensional_metrics_outside_transaction import (
    validate_dimensional_metrics_outside_transaction,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

import newrelic.core.otlp_utils
from newrelic.api.application import application_instance
from newrelic.api.background_task import background_task
from newrelic.api.transaction import (
    record_dimensional_metric,
    record_dimensional_metrics,
)
from newrelic.common.metric_utils import create_metric_identity
from newrelic.core.config import global_settings
from newrelic.packages import six

try:
    # python 2.x
    reload
except NameError:
    # python 3.x
    from importlib import reload


@pytest.fixture(scope="module", autouse=True, params=["protobuf", "json"])
def otlp_content_encoding(request):
    if six.PY2 and request.param == "protobuf":
        pytest.skip("OTLP protos are not compatible with Python 2.")

    _settings = global_settings()
    prev = _settings.debug.otlp_content_encoding
    _settings.debug.otlp_content_encoding = request.param
    reload(newrelic.core.otlp_utils)
    assert newrelic.core.otlp_utils.otlp_content_setting == request.param, "Content encoding mismatch."

    yield

    _settings.debug.otlp_content_encoding = prev


_test_tags_examples = [
    (None, None),
    ({}, None),
    ([], None),
    ({"str": "a"}, frozenset({("str", "a")})),
    ({"int": 1}, frozenset({("int", 1)})),
    ({"float": 1.0}, frozenset({("float", 1.0)})),
    ({"bool": True}, frozenset({("bool", True)})),
    ({"list": [1]}, frozenset({("list", "[1]")})),
    ({"dict": {"subtag": 1}}, frozenset({("dict", "{'subtag': 1}")})),
    ([("tags-as-list", 1)], frozenset({("tags-as-list", 1)})),
]


@pytest.mark.parametrize("tags,expected", _test_tags_examples)
def test_create_metric_identity(tags, expected):
    name = "Metric"
    output_name, output_tags = create_metric_identity(name, tags=tags)
    assert output_name == name, "Name does not match."
    assert output_tags == expected, "Output tags do not match."


@pytest.mark.parametrize("tags,expected", _test_tags_examples)
@reset_core_stats_engine()
def test_record_dimensional_metric_inside_transaction(tags, expected):
    @validate_transaction_metrics(
        "test_record_dimensional_metric_inside_transaction",
        background_task=True,
        dimensional_metrics=[
            ("Metric", expected, 1),
        ],
    )
    @background_task(name="test_record_dimensional_metric_inside_transaction")
    def _test():
        record_dimensional_metric("Metric", 1, tags=tags)

    _test()


@pytest.mark.parametrize("tags,expected", _test_tags_examples)
@reset_core_stats_engine()
def test_record_dimensional_metric_outside_transaction(tags, expected):
    @validate_dimensional_metrics_outside_transaction([("Metric", expected, 1)])
    def _test():
        app = application_instance()
        record_dimensional_metric("Metric", 1, tags=tags, application=app)

    _test()


@pytest.mark.parametrize("tags,expected", _test_tags_examples)
@reset_core_stats_engine()
def test_record_dimensional_metrics_inside_transaction(tags, expected):
    @validate_transaction_metrics(
        "test_record_dimensional_metrics_inside_transaction",
        background_task=True,
        dimensional_metrics=[("Metric.1", expected, 1), ("Metric.2", expected, 1)],
    )
    @background_task(name="test_record_dimensional_metrics_inside_transaction")
    def _test():
        record_dimensional_metrics([("Metric.1", 1, tags), ("Metric.2", 1, tags)])

    _test()


@pytest.mark.parametrize("tags,expected", _test_tags_examples)
@reset_core_stats_engine()
def test_record_dimensional_metrics_outside_transaction(tags, expected):
    @validate_dimensional_metrics_outside_transaction([("Metric.1", expected, 1), ("Metric.2", expected, 1)])
    def _test():
        app = application_instance()
        record_dimensional_metrics([("Metric.1", 1, tags), ("Metric.2", 1, tags)], application=app)

    _test()


@reset_core_stats_engine()
def test_dimensional_metrics_different_tags():
    @validate_transaction_metrics(
        "test_dimensional_metrics_different_tags",
        background_task=True,
        dimensional_metrics=[
            ("Metric", frozenset({("tag", 1)}), 1),
            ("Metric", frozenset({("tag", 2)}), 2),
        ],
    )
    @background_task(name="test_dimensional_metrics_different_tags")
    def _test():
        record_dimensional_metrics(
            [
                ("Metric", 1, {"tag": 1}),
                ("Metric", 1, {"tag": 2}),
            ]
        )
        record_dimensional_metric("Metric", 1, {"tag": 2})

    _test()


@reset_core_stats_engine()
@validate_dimensional_metric_payload(
    summary_metrics=[
        ("Metric.Summary", {"tag": 1}, 1),
        ("Metric.Summary", {"tag": 2}, 1),
        ("Metric.Summary", None, 1),
        ("Metric.Mixed", {"tag": 1}, 1),
        ("Metric.NotPresent", None, None),
    ],
    count_metrics=[
        ("Metric.Count", {"tag": 1}, 1),
        ("Metric.Count", {"tag": 2}, 2),
        ("Metric.Count", None, 3),
        ("Metric.Mixed", {"tag": 2}, 2),
        ("Metric.NotPresent", None, None),
    ],
)
def test_dimensional_metrics_payload():
    @background_task(name="test_dimensional_metric_payload")
    def _test():
        record_dimensional_metrics(
            [
                ("Metric.Summary", 1, {"tag": 1}),
                ("Metric.Summary", 2, {"tag": 2}),
                ("Metric.Summary", 3),  # No tags
                ("Metric.Count", {"count": 1}, {"tag": 1}),
                ("Metric.Count", {"count": 2}, {"tag": 2}),
                ("Metric.Count", {"count": 3}),  # No tags
                ("Metric.Mixed", 1, {"tag": 1}),
                ("Metric.Mixed", {"count": 2}, {"tag": 2}),
            ]
        )

    _test()
    app = application_instance()
    core_app = app._agent.application(app.name)
    core_app.harvest()


@reset_core_stats_engine()
@validate_dimensional_metric_payload(
    summary_metrics=[
        ("Metric.Summary", None, 1),
        ("Metric.Count", None, None),  # Should NOT be present
    ],
    count_metrics=[
        ("Metric.Count", None, 1),
        ("Metric.Summary", None, None),  # Should NOT be present
    ],
)
def test_dimensional_metrics_no_duplicate_encodings():
    @background_task(name="test_dimensional_metric_payload")
    def _test():
        record_dimensional_metrics(
            [
                ("Metric.Summary", 1),
                ("Metric.Count", {"count": 1}),
            ]
        )

    _test()
    app = application_instance()
    core_app = app._agent.application(app.name)
    core_app.harvest()
